import os

############
#Definitions
############

Import('mode')

GLOBAL_DEFINES=[]

use_gcc = GetOption('use_gcc')

compile_hle = GetOption('hle')

avlcatreelock = GetOption('avlcatreelock')

if compile_hle:
    GLOBAL_DEFINES = GLOBAL_DEFINES + [("USE_HLE_LOCK", "1")]


#CATREE_PTHREAD=1
#CATREE_TATAS=2
#CATREE_SEQ=3
#CATREE_HLE=4
if avlcatreelock:
    GLOBAL_DEFINES = GLOBAL_DEFINES + [("CATREE_LOCK_TYPE", avlcatreelock)]

std_cc_flags = ['-std=gnu11',
                '-Wall',
                '-Wextra',
                '-pedantic',
                '-fPIC',
                '-pthread']
std_cxx_flags = ['-std=c++11',
		 '-Wall',
		 '-Wextra',
		 '-pedantic',
		 '-fPIC']

if use_gcc:
    std_cc_flags = std_cc_flags + ['-mrtm']

std_link_flags = ['-pthread']

debug_flags = ['-O1', '-g']#, '-fsanitize=address-full', '-fno-omit-frame-pointer']

optimization_flags = ['-O3', '-g']

profile_flags = ['-fno-omit-frame-pointer -O2']

if(mode=='debug'):
    std_cc_flags = std_cc_flags + debug_flags
    std_cxx_flags = std_cxx_flags + debug_flags
    std_link_flags = std_link_flags + debug_flags
elif(mode=='profile'):
    std_cc_flags = std_cc_flags + profile_flags
    std_cxx_flags = std_cxx_flags + profile_flags
    std_link_flags = std_link_flags + profile_flags
else:
    std_cc_flags = std_cc_flags + optimization_flags
    std_cxx_flags = std_cxx_flags + optimization_flags
    std_link_flags = std_link_flags + optimization_flags

cc = 'clang'
cxx = 'clang++'

if use_gcc:
    cc = 'gcc-4.8'
    cxx = 'g++'

env = Environment(
    ENV = os.environ,
    CFLAGS = ' '.join(std_cc_flags),
    CXXFLAGS = ' '.join(std_cxx_flags),
    CC = cc,
    CXX = cxx,
    LINKFLAGS = ' '.join(std_link_flags),
    CPPPATH = ['src/c/'])


######
#Build
######

read_indicator_object = env.Object(source='src/c/read_indicators/reader_groups_read_indicator.c', CPPDEFINES=GLOBAL_DEFINES)
ccsynch_lock_object = env.Object(source='src/c/locks/ccsynch_lock.c', CPPDEFINES=GLOBAL_DEFINES)
drmcs_lock_object = env.Object(source='src/c/locks/drmcs_lock.c', CPPDEFINES=GLOBAL_DEFINES)
mcs_lock_object = env.Object(source='src/c/locks/mcs_lock.c', CPPDEFINES=GLOBAL_DEFINES)
mrqd_lock_object = env.Object(source='src/c/locks/mrqd_lock.c', CPPDEFINES=GLOBAL_DEFINES)
qd_lock_object = env.Object(source='src/c/locks/qd_lock.c', CPPDEFINES=GLOBAL_DEFINES)
tatas_lock_object = env.Object(source='src/c/locks/tatas_lock.c', CPPDEFINES=GLOBAL_DEFINES)
seq_lock_object = env.Object(source='src/c/locks/seq_lock.c', CPPDEFINES=GLOBAL_DEFINES)
hle_lock_object = None
lock_dependencies = [read_indicator_object,ccsynch_lock_object,drmcs_lock_object,mcs_lock_object,mrqd_lock_object,qd_lock_object,tatas_lock_object,seq_lock_object]

if compile_hle:
    hle_lock_object = env.Object(source='src/c/locks/experimental/hle_lock.c', CPPDEFINES=GLOBAL_DEFINES)
    lock_dependencies = lock_dependencies + hle_lock_object

chained_hash_set_object = env.Object(source='src/c/data_structures/chained_hash_set.c', CPPDEFINES=GLOBAL_DEFINES)
conc_splitch_set_object = env.Object(source='src/c/data_structures/conc_splitch_set.c', CPPDEFINES=GLOBAL_DEFINES)
sorted_list_set_object = env.Object(source='src/c/data_structures/sorted_list_set.c', CPPDEFINES=GLOBAL_DEFINES)
atree_set_object = env.Object(source='src/c/data_structures/atree_set.c', CPPDEFINES=GLOBAL_DEFINES)
avlatree_set_object = env.Object(source='src/c/data_structures/avlatree_set.c', CPPDEFINES=GLOBAL_DEFINES)
treapatree_set_object = env.Object(source='src/c/data_structures/treapatree_set.c', CPPDEFINES=GLOBAL_DEFINES)
ltree_set_object = env.Object(source='src/c/data_structures/ltree_set.c', CPPDEFINES=GLOBAL_DEFINES)

data_structures_dependencies = [chained_hash_set_object,conc_splitch_set_object,sorted_list_set_object, atree_set_object, avlatree_set_object, treapatree_set_object, ltree_set_object]

dependencies = lock_dependencies + data_structures_dependencies


#Static Library
################

static_lib = env.StaticLibrary(target = 'qd_lock_lib', source = dependencies)


#Dynamic Library
################

lib_src = (Glob('src/c/data_structures/*.c') + 
           Glob('src/c/locks/*.c') +
           Glob('src/c/read_indicators/*.c'))

if compile_hle:
    lib_src + Glob('src/c/read_indicators/experimental/hle_lock.c')

dynamic_lib = env.SharedLibrary(target = 'qd_lock_lib',
                                source = lib_src
                                )

#Tests
######

env.Program(source=['src/c/tests/test_lock.c'] + dependencies,
            target='test_lock',
            CPPDEFINES=[('LOCK_TYPE', 'OOLock')]+GLOBAL_DEFINES)

env.Program(source='src/c/tests/test_qd_queue.c',
            target='test_qd_queue')

#Benchmarks
###########

o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
               target='benchmark_set.o',
               CPPDEFINES=[('SET_TYPE', 'OOSet'),
                           ('SET_TYPE_NAME', 'AVLATREE_SET')]+GLOBAL_DEFINES)
env.Program(source=[o]+dependencies,
            target='benchmark_set')

if compile_hle:
    #CATREE
    #WITH HLE
    o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
                target='benchmark_AVLCATREE_HLE.o',
                CPPDEFINES=[('SET_TYPE', 'AVLATreeSet'),
                            ('SET_TYPE_NAME', 'PLAIN_AVLATREE_SET'),
                            ('USE_CA_AVL', '1'),
                            ("CATREE_LOCK_TYPE", "CATREE_HLE")]+GLOBAL_DEFINES)
    env.Program(source=[o]+dependencies,
                target='benchmark_AVLCATREE_HLE')
    #WITHOUT HLE
    o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
                   target='benchmark_AVLCATREE_NOHLE.o',
                   CPPDEFINES=[('SET_TYPE', 'AVLATreeSet'),
                               ('SET_TYPE_NAME', 'PLAIN_AVLATREE_SET'),
                               ('USE_CA_AVL', '1'),
                               ("CATREE_LOCK_TYPE", "CATREE_HLE"),
                               ("NOHLE", "1")]+GLOBAL_DEFINES)
    env.Program(source=[o]+dependencies,
                target='benchmark_AVLCATREE_NOHLE')
    #SINGLE LOCK TREE
    #CATREE
    #WITH HLE
    o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
                target='benchmark_LTREE_HLE.o',
                CPPDEFINES=[('SET_TYPE', 'LTreeSet'),
                            ('SET_TYPE_NAME', 'PLAIN_LTREE_SET'),
                            ('USE_LTREE', '1'),
                            ("LTREE_GLOBAL_LOCK_TYPE", "LTREE_HLE")]+GLOBAL_DEFINES)
    env.Program(source=[o]+dependencies,
                target='benchmark_LTREE_HLE')
    #WITHOUT HLE
    o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
                   target='benchmark_LTREE_NOHLE.o',
                   CPPDEFINES=[('SET_TYPE', 'LTreeSet'),
                               ('SET_TYPE_NAME', 'PLAIN_LTREE_SET'),
                               ('USE_LTREE', '1'),
                               ("LTREE_GLOBAL_LOCK_TYPE", "LTREE_HLE"),
                               ("NOHLE", "1")]+GLOBAL_DEFINES)
    env.Program(source=[o]+dependencies,
                target='benchmark_LTREE_NOHLE')


#CA TREE
o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
               target='benchmark_AVLCATREE_TATAS.o',
               CPPDEFINES=[('SET_TYPE', 'AVLATreeSet'),
                           ('SET_TYPE_NAME', 'PLAIN_AVLATREE_SET'),
                           ('USE_CA_AVL', '1'),
                           ("CATREE_LOCK_TYPE", "CATREE_TATAS")]+GLOBAL_DEFINES)
env.Program(source=[o]+dependencies,
            target='benchmark_AVLCATREE_TATAS')

o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
               target='benchmark_AVLCATREE_PTHREAD.o',
               CPPDEFINES=[('SET_TYPE', 'AVLATreeSet'),
                           ('SET_TYPE_NAME', 'PLAIN_AVLATREE_SET'),
                           ('USE_CA_AVL', '1'),
                           ("CATREE_LOCK_TYPE", "CATREE_PTHREAD")]+GLOBAL_DEFINES)
env.Program(source=[o]+dependencies,
            target='benchmark_AVLCATREE_PTHREAD')

o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
               target='benchmark_AVLCATREE_DRMCS.o',
               CPPDEFINES=[('SET_TYPE', 'AVLATreeSet'),
                           ('SET_TYPE_NAME', 'PLAIN_AVLATREE_SET'),
                           ('USE_CA_AVL', '1'),
                           ("CATREE_LOCK_TYPE", "CATREE_DRMCS")]+GLOBAL_DEFINES)
env.Program(source=[o]+dependencies,
            target='benchmark_AVLCATREE_DRMCS')

o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
               target='benchmark_AVLCATREE_MCS.o',
               CPPDEFINES=[('SET_TYPE', 'AVLATreeSet'),
                           ('SET_TYPE_NAME', 'PLAIN_AVLATREE_SET'),
                           ('USE_CA_AVL', '1'),
                           ("CATREE_LOCK_TYPE", "CATREE_MCS")]+GLOBAL_DEFINES)
env.Program(source=[o]+dependencies,
            target='benchmark_AVLCATREE_MCS')


#SINGLE LOCK TREE
o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
               target='benchmark_LTREE_TATAS.o',
               CPPDEFINES=[('SET_TYPE', 'LTreeSet'),
                           ('SET_TYPE_NAME', 'PLAIN_LTREE_SET'),
                           ('USE_LTREE', '1'),
                           ("LTREE_GLOBAL_LOCK_TYPE", "LTREE_TATAS")]+GLOBAL_DEFINES)
env.Program(source=[o]+dependencies,
            target='benchmark_LTREE_TATAS')

o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
               target='benchmark_LTREE_PTHREAD.o',
               CPPDEFINES=[('SET_TYPE', 'LTreeSet'),
                           ('SET_TYPE_NAME', 'PLAIN_LTREE_SET'),
                           ('USE_LTREE', '1'),
                           ("LTREE_GLOBAL_LOCK_TYPE", "LTREE_PTHREAD")]+GLOBAL_DEFINES)
env.Program(source=[o]+dependencies,
            target='benchmark_LTREE_PTHREAD')

o = env.Object(source=['src/c/benchmarks/benchmark_set.c'],
               target='benchmark_LTREE_DRMCS.o',
               CPPDEFINES=[('SET_TYPE', 'LTreeSet'),
                           ('SET_TYPE_NAME', 'PLAIN_LTREE_SET'),
                           ('USE_LTREE', '1'),
                           ("LTREE_GLOBAL_LOCK_TYPE", "LTREE_DRMCS")]+GLOBAL_DEFINES)
env.Program(source=[o]+dependencies,
            target='benchmark_LTREE_DRMCS')


if not use_gcc:
    #Plain locks
    #type, type name
    all_locks = [('SEQLock', 'PLAIN_SEQ_LOCK'),
                 ('TATASLock', 'PLAIN_TATAS_LOCK'),
                 ('QDLock', 'PLAIN_QD_LOCK'),
                 ('MRQDLock', 'PLAIN_MRQD_LOCK'),
                 ('CCSynchLock', 'PLAIN_CCSYNCH_LOCK'),
                 ('MCSLock', 'PLAIN_MCS_LOCK'),
                 ('DRMCSLock', 'PLAIN_DRMCS_LOCK')]
    

    if compile_hle:
        all_locks = all_locks + [('HLELock', 'PLAIN_HLE_LOCK')]

    for (lock_type, lock_type_name) in all_locks:
        object = env.Object(source='src/c/tests/test_lock.c',
                            target='objects/test_' + lock_type_name + '.o',
                            CPPDEFINES=[('LOCK_TYPE', lock_type),
                                        ('LOCK_TYPE_NAME', lock_type_name)] +GLOBAL_DEFINES)
        env.Program(source=[object] + dependencies,
                    target='test_' + lock_type_name)
    #Plain Sets
    #type, type name
    all_sets = [('ChainedHashSet', 'PLAIN_CHAINED_HASH_SET'),
                ('SortedListSet', 'PLAIN_SORTED_LIST_SET'),
                ('ConcSplitchSet', 'PLAIN_CONC_SPLITCH_SET'),
                ('ATreeSet', 'PLAIN_ATREE_SET'),
                ('AVLATreeSet', 'PLAIN_AVLATREE_SET'),
                ('TREAPATreeSet', 'PLAIN_TREAPATREE_SET'),
                ('LTreeSet', 'PLAIN_LTREE_SET')
                ]
    
    for (set_type, set_type_name) in all_sets:
        #Tests
        object = env.Object(source='src/c/tests/test_set.c',
                            target='objects/test_' + set_type_name + '.o',
                            CPPDEFINES=[('SET_TYPE', set_type),
                                        ('SET_TYPE_NAME', set_type_name)] + GLOBAL_DEFINES)
        env.Program(source=[object] + dependencies,
                    target='test_' + set_type_name)
        #Bechmarks
        bench_object = env.Object(source='src/c/benchmarks/benchmark_set.c',
                                  target='objects/benchmark_' + set_type_name + '.o',
                                  CPPDEFINES=[('SET_TYPE', set_type),
                                              ('SET_TYPE_NAME', set_type_name)] + GLOBAL_DEFINES)
        env.Program(source=[bench_object] + dependencies,
                    target='benchmark_' + set_type_name)
        # LIBPATH = ["/home/kjell/Dropbox/projects/qd_lock_lib/gperftools-2.2.1"],
        # LINKFLAGS="--static",
        # LIBS=["tcmalloc"],


#Examples
#########

env.Program(source=['src/c/examples/qd_lock_delegate_example.c'] + static_lib,
            target='qd_lock_delegate_example',
            CPPDEFINES=GLOBAL_DEFINES)

env.Program(source=['src/c/examples/shared_int_example.c'] + static_lib,
            target='shared_int_example',
            CPPDEFINES=GLOBAL_DEFINES)

env.Program(source=['src/c/examples/concurrent_queue_example.c'] + static_lib,
            target='concurrent_queue_example',
            CPPDEFINES=GLOBAL_DEFINES)


#Copy scripts
#############
Command('run_benchmarks.py', 
        'src/python/run_benchmarks.py', 
        Copy("$TARGET", "$SOURCE"))

Command('compare_benchmarks.py', 
        'src/python/compare_benchmarks.py', 
        Copy("$TARGET", "$SOURCE"))

Command('benchmark_runner.py', 
        'src/python/benchmark_runner.py', 
        Copy("$TARGET", "$SOURCE"))
