
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/time.h>
#include <assert.h>
#include <limits.h>
#include <sched.h>

#include "misc/padded_types.h"

#define AVLATREE_SPECIAL_COPY_OUT \
    long * asLong = ((long*)currentNode->value); \
    void * toReturn = (void*)(*asLong);
#define AVLATREE_SPECIAL_COPY_OUT_PARAMS
#define AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS
#define AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS_NULL


#define LTREE_SPECIAL_COPY_OUT \
    long * asLong = ((long*)currentNode->value); \
    void * toReturn = (void*)(*asLong);
#define LTREE_SPECIAL_COPY_OUT_PARAMS
#define LTREE_SPECIAL_COPY_OUT_PARAMS_PASS
#define LTREE_SPECIAL_COPY_OUT_PARAMS_PASS_NULL


#ifdef USE_CA_AVL
#   define INSERT aat_set_insert
#   define LOOKUP aat_set_lookup
#   define DELETE aat_set_delete
#   define FREE aat_set_free
#   define CREATE(SET_TYPE_NAME,keypos,my_extract_key,my_hash_key,my_are_equal,my_less_than,my_to_string) plain_aat_set_create(keypos,my_extract_key,my_hash_key,my_are_equal,my_less_than,my_to_string)
#elif defined (USE_LTREE)
#   define INSERT lt_set_insert
#   define LOOKUP lt_set_lookup
#   define DELETE lt_set_delete
#   define FREE lt_set_free
#   define CREATE(SET_TYPE_NAME,keypos,my_extract_key,my_hash_key,my_are_equal,my_less_than,my_to_string) plain_lt_set_create(keypos,my_extract_key,my_hash_key,my_are_equal,my_less_than,my_to_string)
#else
#   define INSERT SET_insert
#   define LOOKUP SET_lookup
#   define DELETE SET_delete
#   define FREE SET_free
#   define CREATE SET_create
#endif

#include "data_structures/sets.h"




//Should be power of 2
#define NUMBER_OF_ELEMENTS_IN_ARRAYS 64
#define ELEMENT_POS_MASK (NUMBER_OF_ELEMENTS_IN_ARRAYS-1)
#define ADD_RAND_NUM_MASK (128-1)
#define FALSE_SHARING_SECURITY 128


#ifdef PINNING
_Alignas(CACHE_LINE_SIZE)
__thread CacheLinePaddedInt numa_node;
//#define CORE_FIRST_POLICY_SANDY
//#define SPREAD_PINNING_POLICY
#define NODE_FIRST_POLICY
#endif

typedef struct SeedWrapperImpl{
    char pad1[FALSE_SHARING_SECURITY];
    unsigned short value[3];
    char pad2[FALSE_SHARING_SECURITY - (3*sizeof(short))];
} SeedWrapper;

typedef struct BoolWrapperImpl{
    char pad1[FALSE_SHARING_SECURITY];
    volatile atomic_bool value;
    char pad2[FALSE_SHARING_SECURITY - sizeof(atomic_bool)];
} BoolWrapper;


typedef struct LockThreadLocalSeedImpl{
    int thread_id;
    SET_TYPE * set;
    unsigned short * seed;
} LockThreadLocalSeed;


//=======================
//Benchmark mutable state (Important that they are all on separate cache lines)
//=======================

_Alignas(CACHE_LINE_SIZE)
BoolWrapper benchmarkStopedWrapper;


_Alignas(CACHE_LINE_SIZE)
BoolWrapper benchmarkStartedWrapper;

_Alignas(CACHE_LINE_SIZE)
SeedWrapper threadLocalSeeds[128];

//========================
//Benchmark imutable state
//========================

typedef struct ImutableStateWrapperImpl{
    char pad1[FALSE_SHARING_SECURITY];
    double percentageRead;
    int keyRange;
    int numberOfWarmUpInserts;
    SET_TYPE * set;
    char pad2[FALSE_SHARING_SECURITY];
} ImutableStateWrapper;

_Alignas(CACHE_LINE_SIZE)
ImutableStateWrapper imsw = 
{ 
    .percentageRead = 0.8,
    .keyRange = 2000000,
    .numberOfWarmUpInserts = 1000000,
};


//=============
//The Benchmark
//=============


#ifdef PINNING

int numa_structure[NUMBER_OF_NUMA_NODES][NUMBER_OF_CPUS_PER_NODE] = NUMA_STRUCTURE;
int core_in_node_counters[NUMBER_OF_NUMA_NODES];
int next_numa_node = 0;
bool first = true;
pthread_mutex_t thread_pin_mutex = PTHREAD_MUTEX_INITIALIZER;

void pin(int thread_id){
    pthread_mutex_lock(&thread_pin_mutex);
    if(first){
        first = false;
        for(int i = 0; i < NUMBER_OF_NUMA_NODES; i++){
            core_in_node_counters[i] = 0;
        }
    }


#ifdef NODE_FIRST_POLICY
    if((core_in_node_counters[next_numa_node] != 0) && (0 == (core_in_node_counters[next_numa_node] % NUMBER_OF_CPUS_PER_NODE))){
        next_numa_node++;
    }
#endif
#ifdef CORE_FIRST_POLICY_SANDY
    if((core_in_node_counters[next_numa_node] != 0) && (0 == (core_in_node_counters[next_numa_node] % NUMBER_OF_CPUS_PER_NODE))){
        next_numa_node++;
    }
#endif

    int node = next_numa_node % NUMBER_OF_NUMA_NODES;
    numa_node.value = node;   
    int core_in_node = core_in_node_counters[next_numa_node] % NUMBER_OF_CPUS_PER_NODE;
#ifdef CORE_FIRST_POLICY_SANDY
    if(((core_in_node+1) % 2) == 0){
        core_in_node = NUMBER_OF_CPUS_PER_NODE/2 + ((core_in_node+1)/2) - 1;
    }else if(core_in_node != 0){
        core_in_node = core_in_node / 2;
    }
#endif

    int core_to_pin_to = numa_structure[node][core_in_node]; 

#ifdef SPREAD_PINNING_POLICY
    next_numa_node++;
#endif
    core_in_node_counters[next_numa_node]++;

    int ret = 0;
    cpu_set_t mask;  
    unsigned int len = sizeof(mask);
    CPU_ZERO(&mask);
    CPU_SET(core_to_pin_to, &mask);
    ret = sched_setaffinity(0, len, &mask);
    if (ret == -1){
        printf("sched_setaffinity failed!!\n");
        exit(0);
    }
    pthread_mutex_unlock(&thread_pin_mutex);
}
#endif

int globalDummy = 0; //To avoid that the compiler optimize away read 

void *mixed_read_write_benchmark_thread(void *lockThreadLocalSeedPointer){
    LockThreadLocalSeed * lockThreadLocalSeed = (LockThreadLocalSeed *)lockThreadLocalSeedPointer; 
    unsigned short * xsubi = lockThreadLocalSeed->seed;
    double deleteLimit =  imsw.percentageRead + ((1.0 - imsw.percentageRead) / 2.0);
    long dummy = 0;//To avoid that the compiler optimize away read
    unsigned long totalNumberOfOps = 0;
    SET_TYPE * set = imsw.set;
#ifdef PINNING
    pin(lockThreadLocalSeed->thread_id);
#endif

    //START LINE
    while(!atomic_load_explicit(&benchmarkStartedWrapper.value, memory_order_acquire)){
        atomic_thread_fence(memory_order_seq_cst);
    }
    while(!atomic_load_explicit(&benchmarkStopedWrapper.value, memory_order_acquire)){
        long randomNumber = (jrand48(xsubi)+2147483648 )% imsw.keyRange;
        double randomOp = erand48(xsubi);
        if(randomOp < imsw.percentageRead){
            //Make a random read operation
            void * element = LOOKUP(set, &randomNumber);
            if(element != NULL){
                //printf("%ld\n ", (long)element);
                dummy = dummy + (long)element;
            }
            //printf("LOOKUP\n");
       }else if(randomOp < deleteLimit){
            //Make a ranom insert operation
            //printf("INSERT %ld\n ", randomNumber);
            INSERT(set, &randomNumber, sizeof(int));
        }else {
            //printf("DELETE\n");
            //Make a random delete operation
            DELETE(set, &randomNumber, sizeof(int));
        }
        totalNumberOfOps++;
    }

    globalDummy = dummy;

    return (void*)(totalNumberOfOps);
}

//================
//Benchmark Runner
//================


typedef struct {
    unsigned long ops;
    long time;
} result;



void * my_extract_key(void * v, int keyPos){
    (void)keyPos;
    return v;
}

unsigned int my_hash_key(void * keyPtr){
    int key = *((int *)keyPtr);
    return ((unsigned int)key);
}

bool my_are_equal(void * k1, void * k2){
    int  k1i = *((int *)k1);
    int k2i = *((int *)k2);
    return k1i == k2i;
}

bool my_less_than(void * k1, void * k2){
    int  k1i = *((int *)k1);
    int k2i = *((int *)k2);
    return k1i < k2i;
}

char * my_to_string(void * value){
    int  valuei = *((int *)value);
    char * str = malloc(100);
    sprintf(str, "\"%d\"", valuei);
    return str;
}


result benchmark_parallel_mixed_read_write(double percentageReadParam, 
                                           int numberOfThreads,
                                           int benchmarkTimeSeconds,
                                           int keyRange,
                                           int numberOfWarmUpInserts){
    if(numberOfThreads>128){
        printf("You need to increase the threadLocalSeeds variable!\n");
        assert(false);
    }


    imsw.percentageRead = percentageReadParam;
    imsw.keyRange = keyRange;
    imsw.numberOfWarmUpInserts = numberOfWarmUpInserts;
    atomic_store( &benchmarkStartedWrapper.value, false);
    atomic_store( &benchmarkStopedWrapper.value, false);
    pthread_t threads[numberOfThreads];
    struct timeval timeStart;
    struct timeval timeEnd;

    SET_TYPE * set =  (SET_TYPE *)CREATE(SET_TYPE_NAME,
                                         0,
                                         my_extract_key,
                                         my_hash_key,
                                         my_are_equal,
                                         my_less_than,
                                         my_to_string);

    imsw.set = set;
    srand48(79543);
    for(int i = 0; i < numberOfWarmUpInserts; i++){
        int insertValue = random() % keyRange;
        INSERT(set, &insertValue, sizeof(int));
    }

    LockThreadLocalSeed lockThreadLocalSeeds[numberOfThreads];

    for(int i = 0; i < numberOfThreads; i ++){
        unsigned short * seed = threadLocalSeeds[i].value;
        srand48(i);
        unsigned short * seedResult = seed48(seed);
        seed[0] = seedResult[0];
        seed[1] = seedResult[1];
        seed[2] = seedResult[2];
        LockThreadLocalSeed * lockThreadLocalSeed = &lockThreadLocalSeeds[i];
        lockThreadLocalSeed->seed = seed;
        lockThreadLocalSeed->thread_id = i;
        pthread_create(&threads[i],NULL,&mixed_read_write_benchmark_thread,lockThreadLocalSeed);
    }
    usleep(100000);//To make sure all threads are set up
    gettimeofday(&timeStart, NULL);
    atomic_store_explicit( &benchmarkStartedWrapper.value, true, memory_order_release );
    __sync_synchronize();

    usleep(1000000 * benchmarkTimeSeconds);

    atomic_store_explicit( &benchmarkStopedWrapper.value, true, memory_order_release );
    __sync_synchronize();

    unsigned long totalNumberOfOperations = 0;
    for(int i = 0; i < numberOfThreads; i++){
        unsigned long threadNumberOfOperations;
        pthread_join(threads[i],(void*)&threadNumberOfOperations);
        totalNumberOfOperations = totalNumberOfOperations + threadNumberOfOperations;
    }
    gettimeofday(&timeEnd, NULL);

    FREE(set);

    long benchmarRealTime = (timeEnd.tv_sec-timeStart.tv_sec)*1000000 + timeEnd.tv_usec-timeStart.tv_usec;

    result r;
    r.ops = totalNumberOfOperations;
    r.time = benchmarRealTime;
    return r;
}


void run_scaling_benchmark(char * benchmarkName,
                           int numOfThreads,
                           double percentageReadParam,
                           int benchmarkTimeSeconds,
                           int keyRange,
                           int numberOfWarmUpInserts){
    fprintf(stderr, "=> Benchmark %d threads\n", numOfThreads);
    result r = benchmark_parallel_mixed_read_write(percentageReadParam, 
                                                   numOfThreads,
                                                   benchmarkTimeSeconds,
                                                   keyRange,
                                                   numberOfWarmUpInserts);
    printf("%s %d %ld %lu\n", benchmarkName, numOfThreads, r.time, r.ops);
    fprintf(stderr, "|| %ld microseconds, %lu operations (%d threads)\n", r.time, r.ops, numOfThreads);

}


int main(int argc, char **argv){
    int numberOfStandardArgs = 5;
    if((argc-1) < numberOfStandardArgs){
        printf("The benchmark requires the following paramters:\n");
        printf("\n");
        printf("* Number of threads\n");
        printf("* Percentage read\n");
        printf("* Number of seconds to benchmark\n");
        printf("* Key range\n");
        printf("* Number of Warm up insert operations\n");
        printf("\n");
        printf("Example\n");
        printf("%s 8 0.8 5 2000000 1000000\n", argv[0]);
    }else{
        int numOfThreads = atoi(argv[1]);
        double percentageRead = atof(argv[2]);
        int numberOfSecondsToBenchmark = atoi(argv[3]);
        int keyRange = atoi(argv[4]);
        int numberOfWarmUpInserts = atoi(argv[5]);

        run_scaling_benchmark(argv[0],
                              numOfThreads,
                              percentageRead,
                              numberOfSecondsToBenchmark,
                              keyRange,
                              numberOfWarmUpInserts);
    }
    exit(0);                          
}
