#ifndef SETS_H
#define SETS_H

// Set API
// ========

#include "data_structures/sorted_list_set.h"
#include "data_structures/conc_splitch_set.h"
#include "data_structures/chained_hash_set.h"
#include "data_structures/atree_set.h"
#include "data_structures/avlatree_set.h"
#include "data_structures/treapatree_set.h"
#include "data_structures/ltree_set.h"
#include "data_structures/oo_set_interface.h"
#include "misc/error_help.h"

// ## SET_create

typedef enum {
    LTREE_SET,
    SORTED_LIST_SET,
    CHAINED_HASH_SET,
    CONC_SPLITCH_SET,
    ATREE_SET,
    AVLATREE_SET,
    TREAPATREE_SET,
    PLAIN_LTREE_SET,
    PLAIN_SORTED_LIST_SET,
    PLAIN_CHAINED_HASH_SET,
    PLAIN_CONC_SPLITCH_SET,
    PLAIN_ATREE_SET,
    PLAIN_AVLATREE_SET,
    PLAIN_TREAPATREE_SET
} SET_set_type_name;

void * SET_create(SET_set_type_name setType,
                  unsigned int keyPosition,
                  void * (*extract_key)(void * v, int keyPos),
                  unsigned int (*hash_key)(void * k),
                  bool (*are_equal)(void * v1, void * v2),
                  bool (*less_than)(void * v1, void * v2),
                  char * (*to_string)(void * v1)){
    if(SORTED_LIST_SET == setType){
        return oo_sl_set_create(keyPosition,
                                extract_key,
                                hash_key,
                                are_equal,
                                to_string);
    } else if (CHAINED_HASH_SET == setType){
        return oo_ch_set_create(keyPosition,
                                          extract_key,
                                          hash_key,
                                          are_equal,
                                          to_string);
    } else if (CONC_SPLITCH_SET == setType){
        return oo_csh_set_create(keyPosition,
                                 extract_key,
                                 hash_key,
                                 are_equal,
                                 to_string);
    } else if (LTREE_SET == setType){
        return oo_lt_set_create(keyPosition,
                                extract_key,
                                hash_key,
                                are_equal,
                                less_than,
                                to_string);
    } else if (ATREE_SET == setType){
        return oo_at_set_create(keyPosition,
                                extract_key,
                                hash_key,
                                are_equal,
                                less_than,
                                to_string);
    } else if (AVLATREE_SET == setType){
        return oo_aat_set_create(keyPosition,
                                 extract_key,
                                 hash_key,
                                 are_equal,
                                 less_than,
                                 to_string);
    } else if (TREAPATREE_SET == setType){
        return oo_tat_set_create(keyPosition,
                                 extract_key,
                                 hash_key,
                                 are_equal,
                                 less_than,
                                 to_string);
    }
#ifdef __clang__
    else if(PLAIN_SORTED_LIST_SET == setType){
        return plain_sl_set_create(keyPosition,
                                   extract_key,
                                   hash_key,
                                   are_equal,
                                   to_string);
    } else if (PLAIN_CHAINED_HASH_SET == setType){
        return plain_ch_set_create(keyPosition,
                                   extract_key,
                                   hash_key,
                                   are_equal,
                                   to_string);
    } else if (PLAIN_CONC_SPLITCH_SET == setType){
        return plain_csh_set_create(keyPosition,
                                    extract_key,
                                    hash_key,
                                    are_equal,
                                    to_string);
    } else if (PLAIN_LTREE_SET == setType){
        return plain_lt_set_create(keyPosition,
                                   extract_key,
                                   hash_key,
                                   are_equal,
                                   less_than,
                                   to_string);
    } else if (PLAIN_ATREE_SET == setType){
        return plain_at_set_create(keyPosition,
                                   extract_key,
                                   hash_key,
                                   are_equal,
                                   less_than,
                                   to_string);
    } else if (PLAIN_AVLATREE_SET == setType){
        return plain_aat_set_create(keyPosition,
                                    extract_key,
                                    hash_key,
                                    are_equal,
                                    less_than,
                                    to_string);
    } else if (PLAIN_TREAPATREE_SET == setType){
        return plain_tat_set_create(keyPosition,
                                    extract_key,
                                    hash_key,
                                    are_equal,
                                    less_than,
                                    to_string);
    }
#endif
    LL_error_and_exit("Set type not supported with GCC\n");
    return NULL;/* Should not be reachable */
}

// ## SET_is_concurrent

#ifdef __clang__
#define SET_is_concurrent(X) _Generic((X),\
                             OOSet * : ((OOSet *)X)->m->is_concurrent(((OOSet *)X)->set), \
                             ChainedHashSet * : false,       \
                             SortedListSet * : false, \
                             ConcSplitchSet *: true, \
                             ATreeSet *: true, \
                             LTreeSet *: true, \
                             AVLATreeSet *: true,      \
                             TREAPATreeSet *: true \
                )
#else
#define SET_is_concurrent(X) ((OOSet *)X)->m->is_concurrent(((OOSet *)X)->set)
#endif


// ## SET_free

#ifdef __clang__
#define SET_free(X) _Generic((X),\
                             OOSet * : ooset_free((OOSet*)X), \
                             ChainedHashSet * : ch_set_free((ChainedHashSet *)X),       \
                             SortedListSet * : sl_set_free((SortedListSet *)X), \
                             ConcSplitchSet *: csh_set_free((ConcSplitchSet *)X), \
                             ATreeSet *: at_set_free((ATreeSet *)X),    \
                             LTreeSet *: lt_set_free((LTreeSet *)X),    \
                             AVLATreeSet *: aat_set_free((AVLATreeSet *)X), \
                             TREAPATreeSet *: tat_set_free((TREAPATreeSet *)X)       \
                )
#else
#define SET_free(X) ooset_free(X)
#endif

// ## SET_insert

#ifdef __clang__
#define SET_insert(X, e, size) _Generic((X), \
    OOSet * : ((OOSet *)X)->m->insert(((OOSet *)X)->set, e, size), \
    ChainedHashSet * : ch_set_insert(X, e, size), \
    SortedListSet * : sl_set_insert(X, e, size), \
    ConcSplitchSet *: csh_set_insert((ConcSplitchSet *)X, e, size), \
    ATreeSet *: at_set_insert((ATreeSet *)X, e, size), \
    LTreeSet *: lt_set_insert((LTreeSet *)X, e, size), \
    AVLATreeSet *: aat_set_insert((AVLATreeSet *)X, e, size), \
    TREAPATreeSet *: tat_set_insert((AVLATreeSet *)X, e, size) \
                            )
#else
#define SET_insert(X, e, size) ((OOSet *)X)->m->insert(((OOSet *)X)->set, e, size)
#endif

// ## SET_insert_new

#ifdef __clang__
#define SET_insert_new(X, e, size) _Generic((X), \
    OOSet * : ((OOSet *)X)->m->insert_new(((OOSet *)X)->set, e, size), \
    ChainedHashSet * : ch_set_insert_new(X, e, size), \
    SortedListSet * : sl_set_insert_new(X, e, size), \
    ConcSplitchSet *: csh_set_insert_new((ConcSplitchSet *)X, e, size), \
    ATreeSet *: at_set_insert_new((ATreeSet *)X, e, size), \
    LTreeSet *: lt_set_insert_new((LTreeSet *)X, e, size), \
    AVLATreeSet *: aat_set_insert_new((AVLATreeSet *)X, e, size), \
    TREAPATreeSet *: tat_set_insert_new((TREAPATreeSet *)X, e, size) \
                          )
#else
#define SET_insert_new(X) ((OOSet *)X)->m->insert_new(((OOSet *)X)->set, e, size)
#endif


// ## SET_delete

#ifdef __clang__
#define SET_delete(X, k, kSize) _Generic((X),      \
    OOSet * : ((OOSet *)X)->m->delete(((OOSet *)X)->set, k, kSize), \
    ChainedHashSet * : ch_set_delete(X, k, kSize), \
    SortedListSet * : sl_set_delete(X, k, kSize), \
    ConcSplitchSet *: csh_set_delete((ConcSplitchSet *)X, k, kSize), \
    ATreeSet *: at_set_delete((ATreeSet *)X, k, kSize), \
    LTreeSet *: lt_set_delete((LTreeSet *)X, k, kSize), \
    AVLATreeSet *: aat_set_delete((AVLATreeSet *)X, k, kSize), \
    TREAPATreeSet *: tat_set_delete((TREAPATreeSet *)X, k, kSize) \
                            )
#else
#define SET_delete(X, k, kSize) ((OOSet *)X)->m->delete(((OOSet *)X)->set, k, kSize)
#endif

// ## SET_lookup

#ifdef __clang__
#define SET_lookup(X, k) _Generic((X), \
    OOSet * : ((OOSet *)X)->m->lookup(((OOSet *)X)->set, k), \
    ChainedHashSet * : ch_set_lookup(X, k), \
    SortedListSet * : sl_set_lookup(X, k), \
    ConcSplitchSet *: csh_set_lookup((ConcSplitchSet *)X, k), \
    ATreeSet *: at_set_lookup((ATreeSet *)X, k), \
    LTreeSet *: lt_set_lookup((LTreeSet *)X, k), \
    AVLATreeSet *: aat_set_lookup((AVLATreeSet *)X, k), \
    TREAPATreeSet *: tat_set_lookup((TREAPATreeSet *)X, k) \
                            )
#else
#define SET_lookup(X, k) ((OOSet *)X)->m->lookup(((OOSet *)X)->set, k)
#endif

// ## SET_to_string

#ifdef __clang__
#define SET_to_string(X) _Generic((X), \
    OOSet * : ((OOSet *)X)->m->to_string(((OOSet *)X)->set), \
    ChainedHashSet * : ch_set_to_string(X), \
    SortedListSet * : sl_set_to_string(X), \
    ConcSplitchSet * : csh_set_to_string(X), \
    ATreeSet * : at_set_to_string(X),     \
    LTreeSet * : lt_set_to_string(X),     \
    AVLATreeSet * : aat_set_to_string(X), \
    TREAPATreeSet * : tat_set_to_string(X) \
                            )
#else
#define SET_to_string(X) ((OOSet *)X)->m->to_string(((OOSet *)X)->set)
#endif

// ## SET_print

#ifdef __clang__
#define SET_print(X) _Generic((X), \
    OOSet * : ((OOSet *)X)->m->print(((OOSet *)X)->set), \
    ChainedHashSet * : ch_set_print(X), \
    SortedListSet * : sl_set_print(X),        \
    ConcSplitchSet * : csh_set_print(X),      \
    ATreeSet * : at_set_print(X),     \
    LTreeSet * : lt_set_print(X),     \
    AVLATreeSet * : aat_set_print(X), \
    TREAPATreeSet * : tat_set_print(X)      \
                            )
#else
#define SET_print(X) ((OOSet *)X)->m->print(((OOSet *)X)->set)
#endif


#endif
