#ifndef TREAPATREE_SET_H
#define TREAPATREE_SET_H

/*
 * 
 */


#include <limits.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "misc/thread_includes.h"
#include "misc/padded_types.h"
#include "data_structures/oo_set_interface.h"
#include "locks/drmcs_lock.h"
#include "locks/tatas_lock.h"

#ifndef TREAPATREE_MALLOC
#      define TREAPATREE_MALLOC(size) malloc(size)
#endif

#ifndef TREAPATREE_FREE
#      define TREAPATREE_FREE(data) free(data)
#endif

#ifndef TREAPATREE_NEW_VALUE_NODE
#      define TREAPATREE_NEW_VALUE_NODE(size, value, context) treapAllocateValueNode(size, value)
#endif

#ifndef TREAPATREE_REPLACE_VALUE_NODE
#      define TREAPATREE_REPLACE_VALUE_NODE(size, oldValue, value, custom_free, context) treapReplaceValueNode(size,value,oldValue,custom_free,freeContext)
#endif

#ifndef TREAPATREE_EXTRACT_KEY_UNTRANSFORMED
#      define TREAPATREE_EXTRACT_KEY_UNTRANSFORMED(untransformed) extract_key(valuePtr, keyPosition)
#endif


#define TREAPATREE_MAX_CONTENTION_STATISTICS 1000
#define TREAPATREE_MIN_CONTENTION_STATISTICS -1000
#define TREAPATREE_LOCK_SUCCESS_STATS_CONTRIB -1
#define TREAPATREE_LOCK_FAILURE_STATS_CONTRIB 250
#define TREAPATREE_REBUILD_CONSTANT 2
#define TREAPATREE_DEFAULT_STACK_SIZE 50
#define TREAPATREE_LEFT_DIR 0
#define TREAPATREE_RIGHT_DIR 1
#define TREAPATREE_ROOT_DIR 2 
#define TREAPATREE_FREE_LIST_SIZE 1024 
//#define TREAPATREE_COUNT_ROUTE_NODES 1

//#define TREAPATREE_PTHREAD_LOCK 1
#define TREAPATREE_TATAS_LOCK 1

#ifdef TREAPATREE_PTHREAD_LOCK
#define TREAPATREE_LOCK_TYPE pthread_mutex_t
#define TREAPATREE_LOCK_INITIALIZER(lock_ptr)                           \
    pthread_mutex_init(lock_ptr, NULL);

   /* pthread_mutexattr_t attr;                                         \ */
   /* pthread_mutexattr_init(&attr);                                    \ */
   /* pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);       \ */
   /* pthread_mutexattr_destroy(&attr); */
#define TREAPATREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) (!pthread_mutex_trylock(lock_ptr))
#define TREAPATREE_LOCK_LOCK_FUNCTION(lock_ptr) pthread_mutex_lock(lock_ptr)
#define TREAPATREE_LOCK_UNLOCK_FUNCTION(lock_ptr) pthread_mutex_unlock(lock_ptr)
#define TREAPATREE_LOCK_DESTROY_FUNCTION(lock_ptr) pthread_mutex_destroy(lock_ptr)
#endif


#ifdef TREAPATREE_TATAS_LOCK
#define TREAPATREE_LOCK_TYPE TATASLock
#define TREAPATREE_LOCK_INITIALIZER(lock_ptr) tatas_initialize(lock_ptr);
#define TREAPATREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) tatas_try_lock(lock_ptr)
#define TREAPATREE_LOCK_LOCK_FUNCTION(lock_ptr) tatas_lock(lock_ptr)
#define TREAPATREE_LOCK_UNLOCK_FUNCTION(lock_ptr) tatas_unlock(lock_ptr)
#define TREAPATREE_LOCK_DESTROY_FUNCTION(lock_ptr) ;
#endif

typedef struct {
    TREAPATREE_LOCK_TYPE lock;
    //    int validation;
    int statistics;
} TreapatreeLock;

typedef struct TREAPValueNodeImpl{
    struct TREAPValueNodeImpl * left;
    struct TREAPValueNodeImpl * right;
    long int priority;
    unsigned int valueSize;
    char value[];
} TREAPValueNode;

typedef struct {
    struct TREAPValueNodeImpl * root;
    TreapatreeLock lock;
    bool valid;
} TREAPBaseNode;

typedef struct {
    void * left;
    void * right;
    TATASLock * lock;
    unsigned int valueSize;
    bool valid;
    char value[1];
} TREAPRouteNode;

typedef union {
    TREAPRouteNode route;
    TREAPBaseNode base;
} TREAPBaseOrTREAPRouteNodeContainer;

typedef struct {
    bool isTREAPBaseNode;
    TREAPBaseOrTREAPRouteNodeContainer baseOrRoute;
} TREAPBaseOrTREAPRouteNode;

typedef struct {
    DRMCSLock globalLock;
    TREAPBaseOrTREAPRouteNode * root;
    unsigned int keyPosition;
    void * (*extract_key)(void * v, int keyPos);
    unsigned int (*hash_key)(void * k);
    bool (*are_equal)(void * v1, void * v2);
    bool (*less_than)(void * v1, void * v2);
    char * (*to_string)(void * v1);
    void (*free)(void * v, void * context);
    void * freeContext;
#ifdef TREAPATREE_COUNT_ROUTE_NODES
    char pad1[CACHE_LINE_SIZE*2];
    volatile atomic_int nrOfRouteNodes;
#endif
    char pad[CACHE_LINE_SIZE*2];
    TATASLock freeingFreeListLock;
    volatile atomic_intptr_t freeList[TREAPATREE_FREE_LIST_SIZE];
    LLPaddedULong freeListIndexCounter;
} TREAPATreeSet;


typedef struct {
    char pad1[CACHE_LINE_SIZE];
    bool seedSet;
    unsigned short seed[3];
    char pad2[CACHE_LINE_SIZE*2 - sizeof(bool) - (3*sizeof(unsigned short))];
} TreapThreadLocalSeed;

extern
_Alignas(CACHE_LINE_SIZE)
__thread TreapThreadLocalSeed treapThreadLocalSeed;

extern
_Alignas(CACHE_LINE_SIZE)
OOSetMethodTable TREAPATREE_SET_METHOD_TABLE;

extern
pthread_mutex_t treapATreeSeedInitMutex;


static void enqueueInFreeList(TREAPATreeSet * set, TREAPBaseOrTREAPRouteNode * toEnqueue){
    while(true){
        unsigned long slot = atomic_fetch_add(&set->freeListIndexCounter.value, 1);
        if(slot == TREAPATREE_FREE_LIST_SIZE){
            tatas_lock(&set->freeingFreeListLock);
            if (atomic_load(&set->freeListIndexCounter.value) > TREAPATREE_FREE_LIST_SIZE){//TODO this is probably not needed
                // We are responsible for doing the freeing
                // Ensure that no one still is accessing old data
                drmcs_runlock(&set->globalLock);
                drmcs_lock(&set->globalLock);
                drmcs_unlock(&set->globalLock);
                drmcs_rlock(&set->globalLock);
                for(int i = 0; i < TREAPATREE_FREE_LIST_SIZE; i++){
                    TREAPBaseOrTREAPRouteNode * elementToFree;
                    while((elementToFree = (TREAPBaseOrTREAPRouteNode *)atomic_load(&set->freeList[i])) == NULL){
                        thread_yield();
                    }
                    if(elementToFree->isTREAPBaseNode){
                        TREAPATREE_LOCK_DESTROY_FUNCTION(&elementToFree->baseOrRoute.base.lock.lock); //Free deleted base
                        TREAPATREE_FREE(elementToFree);
                    }else{
                        free(elementToFree->baseOrRoute.route.lock); //Free deleted route
                        TREAPATREE_FREE(elementToFree);   
                    }
                    atomic_store(&set->freeList[i], (intptr_t)NULL);
                }
                //Enqueue our element
                atomic_store(&set->freeList[0], (intptr_t)toEnqueue);
                //Let other threads start to enqueue
                atomic_store(&set->freeListIndexCounter.value, 1);
                tatas_unlock(&set->freeingFreeListLock);
                return;
            }else{
                tatas_unlock(&set->freeingFreeListLock);
                //Retry
            }
        }else if (slot > TREAPATREE_FREE_LIST_SIZE){
            //We need to wait then retry
            drmcs_runlock(&set->globalLock);
            while(atomic_load(&set->freeListIndexCounter.value) > TREAPATREE_FREE_LIST_SIZE){
                thread_yield();
            }
            drmcs_rlock(&set->globalLock);
        } else {
            // Just write our node to our slot
            atomic_store(&set->freeList[slot], (intptr_t)toEnqueue);
            return;
        }
    }
}

static
void init_random_seed_thread(){
    pthread_mutex_lock(&treapATreeSeedInitMutex);
    unsigned short * seed = treapThreadLocalSeed.seed;
    unsigned short * seedResult;
    srand48(pthread_self());
    seedResult = seed48(seed);
    seed[0] = seedResult[0];
    seed[1] = seedResult[1];
    seed[2] = seedResult[2];
    treapThreadLocalSeed.seedSet = true;
    pthread_mutex_unlock(&treapATreeSeedInitMutex);
}

static inline
long int treap_random_number(){
    if(!treapThreadLocalSeed.seedSet){
        init_random_seed_thread();
    }
    return jrand48(treapThreadLocalSeed.seed);
}



static inline
void treapatree_lock_init(TreapatreeLock * lock){
    lock->statistics = 0;
    //    lock->validation = 0;
    TREAPATREE_LOCK_INITIALIZER(&lock->lock);
}

static inline
void treapatree_lock(TreapatreeLock * lock){
    if(TREAPATREE_LOCK_TRY_LOCK_FUNCTION(&lock->lock)){
        lock->statistics += TREAPATREE_LOCK_SUCCESS_STATS_CONTRIB;
        return;
    }
    TREAPATREE_LOCK_LOCK_FUNCTION(&lock->lock);
    lock->statistics += TREAPATREE_LOCK_FAILURE_STATS_CONTRIB;
}


static inline
bool treapatree_trylock(TreapatreeLock * lock){
    if(TREAPATREE_LOCK_TRY_LOCK_FUNCTION(&lock->lock)){
        /* lock->validation++; */
        /* assert(lock->validation == 1); */
        return true;
    }else{
        return false;
    }
}


static inline
void treapatree_unlock(TreapatreeLock * lock){
    /* lock->validation--; */
    /* assert(lock->validation == 0); */
    TREAPATREE_LOCK_UNLOCK_FUNCTION(&lock->lock);
}

static inline
TREAPBaseOrTREAPRouteNode * allocateTREAPBaseNode(){
    TREAPBaseOrTREAPRouteNode * newTREAPBase = TREAPATREE_MALLOC(sizeof(TREAPBaseOrTREAPRouteNode));
    newTREAPBase->isTREAPBaseNode = true;
    treapatree_lock_init(&newTREAPBase->baseOrRoute.base.lock);
    newTREAPBase->baseOrRoute.base.root = NULL;
    newTREAPBase->baseOrRoute.base.valid = true;
    return newTREAPBase;
}

static inline
TREAPBaseOrTREAPRouteNode * treapAllocateRouteNode(unsigned int valueSize,
                                             void * value,
                                             TREAPBaseOrTREAPRouteNode * newTREAPBaseLeft,
                                             TREAPBaseOrTREAPRouteNode * newTREAPBaseRight){
    TREAPBaseOrTREAPRouteNode * newRoute = 
        TREAPATREE_MALLOC(sizeof(TREAPBaseOrTREAPRouteNode) + valueSize - sizeof(char));
    newRoute->isTREAPBaseNode = false;
    newRoute->baseOrRoute.route.lock = plain_tatas_create();
    newRoute->baseOrRoute.route.valid = true;
    newRoute->baseOrRoute.route.left = newTREAPBaseLeft;
    newRoute->baseOrRoute.route.right = newTREAPBaseRight;
    newRoute->baseOrRoute.route.valueSize = valueSize;
    memcpy(newRoute->baseOrRoute.route.value, value, valueSize);
    return newRoute;
}



static inline
TREAPValueNode * treapAllocateValueNode(unsigned int size, void * value){
    TREAPValueNode * newTREAPValueNode = TREAPATREE_MALLOC(sizeof(TREAPValueNode) + size);
    newTREAPValueNode->left = NULL;
    newTREAPValueNode->right = NULL;
    newTREAPValueNode->valueSize = size;
    memcpy(newTREAPValueNode->value, value, size);
    return newTREAPValueNode;
}


static inline
TREAPValueNode * treapReplaceValueNode(unsigned int size,
                                   void * value,
                                   TREAPValueNode * toReplace,
                                   void (*custom_free)(void * v, void * context),
                                   void * freeContext){
    if(toReplace->valueSize < size){
        TREAPValueNode * toReturn = treapAllocateValueNode(size, value);
        custom_free(toReplace, freeContext);
        return toReturn;
    }else{
        toReplace->left = NULL;
        toReplace->right = NULL;
        toReplace->valueSize = size;
        memcpy(toReplace->value, value, size);
        return toReplace;
    }

}

  /* private var root:Node = TREAPBaseNode(null) */

  /* def printRootStructure() = printStructure(root) */

  /* def printStructure(currentNode:Node):String = currentNode match { */
  /*   case TREAPBaseNode(root, _, _, nrOfElements) => "B("+ (if(root == null) null else(root.key)) + ","+nrOfElements+")" */
  /*   case TREAPRouteNode(key, left, right) => "R("+key+","+printStructure(left)+","+printStructure(right)+")" */
  /* } */


//http://www.hackersdelight.org/hdcodetxt/nlz.c.txt
static inline
unsigned int treapNlz1(unsigned int x) {
   int n;

   if (x == 0) return(32);
   n = 0;
   if (x <= 0x0000FFFF) {n = n +16; x = x <<16;}
   if (x <= 0x00FFFFFF) {n = n + 8; x = x << 8;}
   if (x <= 0x0FFFFFFF) {n = n + 4; x = x << 4;}
   if (x <= 0x3FFFFFFF) {n = n + 2; x = x << 2;}
   if (x <= 0x7FFFFFFF) {n = n + 1;}
   return n;
}

static inline
unsigned int treapLog2nlz(unsigned int bits) {
    if( bits == 0 ) {
        return 0; 
    }else {
        return 31 - treapNlz1( bits );
    }
}

 static inline void tat_standard_free(void * toFree, void * freeContext){
     (void)freeContext;
     TREAPATREE_FREE(toFree);
 }

static inline
void plain_tat_set_init(TREAPATreeSet * set,
                       unsigned int keyPosition,
                       void * (*extract_key)(void * v, int keyPos),
                       unsigned int (*hash_key)(void * k),
                       bool (*are_equal)(void * v1, void * v2),
                       bool (*less_than)(void * v1, void * v2),
                       char * (*to_string)(void * v1),
                       void (*custom_free)(void * v, void * context),
                       void * freeContext){
    drmcs_initialize(&set->globalLock);
    set->keyPosition = keyPosition;
    set->extract_key = extract_key;
    set->hash_key = hash_key;
    set->are_equal = are_equal;
    set->less_than = less_than;
    set->to_string = to_string;
    if(custom_free == NULL){
        set->free = tat_standard_free;
        set->freeContext = NULL;
    }else{
        set->free = custom_free;
        set->freeContext = freeContext;
    }
    set->root = allocateTREAPBaseNode();
    tatas_initialize(&set->freeingFreeListLock);
    for(int i = 0; i < TREAPATREE_FREE_LIST_SIZE; i++){
        atomic_init(&set->freeList[i], (intptr_t)NULL);
    }
    atomic_init(&set->freeListIndexCounter.value, 0);
#ifdef TREAPATREE_COUNT_ROUTE_NODES
    atomic_init(&set->nrOfRouteNodes, 0);
#endif
}


static inline
TREAPATreeSet * plain_tat_set_create(unsigned int keyPosition,
                               void * (*extract_key)(void * v, int keyPos),
                               unsigned int (*hash_key)(void * k),
                               bool (*are_equal)(void * v1, void * v2),
                               bool (*less_than)(void * v1, void * v2),
                               char * (*to_string)(void * v1)){
    TREAPATreeSet * set = TREAPATREE_MALLOC(sizeof(TREAPATreeSet));
    plain_tat_set_init(set,
                      keyPosition,
                      extract_key,
                      hash_key,
                      are_equal,
                      less_than,
                      to_string,
                      NULL,
                      NULL);
    return set;
}

static inline
int treap_compare_hash_codes2(unsigned int code1, unsigned int code2){
    return code1 - code2;
}

static inline
unsigned int treapCountNodes(TREAPValueNode * node){ 
    if(node == NULL) return 0;
    else if(node->left == NULL ) return treapCountNodes(node->right) + 1;
    else if(node->right == NULL ) return treapCountNodes(node->left) + 1;
    else return treapCountNodes(node->left) + 1 + treapCountNodes(node->right);
}


static inline
TREAPValueNode * treapInsertTREAPValueNode(TREAPValueNode * insertTo,
                                           TREAPValueNode * insertValueNode,
                                           int keyPosition,
                                           void * (*extract_key)(void * v, int keyPos),
                                           bool (*less_than)(void * v1, void * v2),
                                           TREAPValueNode *** pointersStackParam,
                                           int * dirStackParam,
                                           int stackSizeParam){
    int stackPos = 0;
    TREAPValueNode *** pointersStack;
    int * dirStack;
    int stackSize;
    TREAPValueNode * base = insertTo;
    TREAPValueNode **currentNodePointer = &base;
    void * key;
    TREAPValueNode ** pointersStackS[TREAPATREE_DEFAULT_STACK_SIZE];
    int dirStackS[TREAPATREE_DEFAULT_STACK_SIZE];
    insertValueNode->priority = treap_random_number();
    if(stackSizeParam == 0){
        stackSize = TREAPATREE_DEFAULT_STACK_SIZE;
        pointersStack = pointersStackS;
        dirStack = dirStackS;
    }else{
        stackSize = stackSizeParam;
        pointersStack = pointersStackParam;
        dirStack = dirStackParam;
    }

    key = extract_key(insertValueNode->value, keyPosition);

    dirStack[0] = TREAPATREE_ROOT_DIR;
    pointersStack[0] = currentNodePointer;
    while (true) {
	if(stackPos == (stackSize-1)){
            if(stackSizeParam != 0){
                free(dirStack);
                free(pointersStack);
            }
            stackSize = 2 * stackSize;
            pointersStack = malloc(stackSize * sizeof(TREAPValueNode **));
            dirStack = malloc(stackSize * sizeof(int));
            return treapInsertTREAPValueNode(insertTo,
                                             insertValueNode,
                                             keyPosition,
                                             extract_key,
                                             less_than,
                                             pointersStack,
                                             dirStack,
                                             stackSize);
        }else if (*currentNodePointer == NULL) {
            *currentNodePointer = insertValueNode;
	    (*currentNodePointer)->priority = treap_random_number();
	    (*currentNodePointer)->left = NULL;
	    (*currentNodePointer)->right = NULL;
	    break;
	} else if (less_than(key, extract_key((*currentNodePointer)->value, keyPosition))) {
	    /* left */
            stackPos = stackPos + 1;
	    dirStack[stackPos] = TREAPATREE_LEFT_DIR;
	    currentNodePointer = &((*currentNodePointer)->left);
	    pointersStack[stackPos] = currentNodePointer;
	} else { 
            /* right */
            stackPos = stackPos + 1;
	    dirStack[stackPos] = TREAPATREE_RIGHT_DIR;
	    currentNodePointer = &((*currentNodePointer)->right);
	    pointersStack[stackPos] = currentNodePointer;
	}
    }
    stackPos = stackPos -1;//Now points to the pointer to the parent of the inserted node
    TREAPValueNode * node = (*currentNodePointer);
    while (stackPos >= 0 && (node->priority) > (*pointersStack[stackPos])->priority) {
        if(dirStack[stackPos + 1] == TREAPATREE_LEFT_DIR){
            TREAPValueNode * rotateUp = *pointersStack[stackPos];
            *pointersStack[stackPos] = node;
            rotateUp->left = node->right;
            node->right = rotateUp;
        } else {
            TREAPValueNode * rotateUp = *pointersStack[stackPos];
            *pointersStack[stackPos] = node;
            rotateUp->right = node->left;
            node->left = rotateUp;
        } 
        stackPos = stackPos - 1;
    }
    if(stackSizeParam != 0){
        free(dirStack);
        free(pointersStack);
    }
    return base; 
}

static inline TREAPBaseOrTREAPRouteNode * findParentOf(TREAPBaseOrTREAPRouteNode * container,
                                                       void * value,
                                                       TREAPATreeSet * set,
                                                       int keyPosition,
                                                       void * (*extract_key)(void * v, int keyPos),
                                                       bool (*less_than)(void * v1, void * v2)){
    if(set->root == container){
        return NULL;
    }
    TREAPBaseOrTREAPRouteNode * currentNode = set->root;
    TREAPBaseOrTREAPRouteNode * prevNode = NULL;
    void * key = extract_key(value, keyPosition);
    while(currentNode != container){
        void * nodeKey = extract_key(currentNode->baseOrRoute.route.value, keyPosition);
        prevNode = currentNode;
        if(less_than(key, nodeKey)) {
            currentNode = currentNode->baseOrRoute.route.left;
        } else {
            currentNode = currentNode->baseOrRoute.route.right;
        }
    }
    return prevNode;
}


static inline TREAPBaseOrTREAPRouteNode ** findParentPtrOf(TREAPBaseOrTREAPRouteNode * container,
                                                           void * value,
                                      TREAPATreeSet * set,
                                      int keyPosition,
                                      void * (*extract_key)(void * v, int keyPos),
                                      bool (*less_than)(void * v1, void * v2)){
    TREAPBaseOrTREAPRouteNode * parent = findParentOf(container,
                                                      value,
                                                      set,
                                                      keyPosition,
                                                      extract_key,
                                                      less_than);
    if(parent == NULL){
        return &set->root;
    }
    if(parent->baseOrRoute.route.left == container){
        return (TREAPBaseOrTREAPRouteNode **)&parent->baseOrRoute.route.left;
    }else{
        return (TREAPBaseOrTREAPRouteNode **)&parent->baseOrRoute.route.right;
    }

}


static inline
bool treapContentionReduceSplit(TREAPBaseOrTREAPRouteNode *  baseNodeContainer,
                                void ** parent,
                                TREAPATreeSet * set,
                                int keyPosition,
                                void * (*extract_key)(void * v, int keyPos),
                                bool (*less_than)(void * v1, void * v2)) {
    TREAPBaseNode * node = &baseNodeContainer->baseOrRoute.base;
    if(node == NULL || node->root == NULL || (node->root->left == NULL && node->root->right == NULL)){
        node->lock.statistics = 0;
        return false;
    }
    TREAPValueNode * splitTREAPValueNode = NULL;
      if (node->root->left == NULL) {
          //To get non empty split
          TREAPValueNode * insertTo = node->root->right;
          node->root->right = NULL;
          splitTREAPValueNode = treapInsertTREAPValueNode(insertTo,
                                                          node->root,
                                                          keyPosition,
                                                          extract_key,
                                                          less_than,
                                                          NULL,
                                                          NULL,
                                                          0);
      }else{
          splitTREAPValueNode = node->root;
      }

      void * splitKey = splitTREAPValueNode->value;
      unsigned int splitKeyValueSize = splitTREAPValueNode->valueSize;
      TREAPValueNode * leftTREAPBaseRoot = splitTREAPValueNode->left;
      splitTREAPValueNode->left = NULL;
      TREAPValueNode * rightTREAPBaseRoot = NULL;
      if(splitTREAPValueNode->right == NULL)
          rightTREAPBaseRoot = splitTREAPValueNode;
      else {
          //To narrow down faster
          TREAPValueNode * base = splitTREAPValueNode->right;
          splitTREAPValueNode->right = NULL;
          rightTREAPBaseRoot = treapInsertTREAPValueNode(base,
                                                         splitTREAPValueNode,
                                                         keyPosition,
                                                         extract_key,
                                                         less_than,
                                                         NULL,
                                                         NULL,
                                                         0);
      }
      //We can not use the same because it might contain less than what a thread waiting for expect
      TREAPBaseOrTREAPRouteNode * newTREAPBaseRight = allocateTREAPBaseNode();
      TREAPBaseNode * rightBCont = &newTREAPBaseRight->baseOrRoute.base;
      rightBCont->root = rightTREAPBaseRoot;
 
      TREAPBaseOrTREAPRouteNode * newTREAPBaseLeft = allocateTREAPBaseNode();
      TREAPBaseNode * leftBCont = &newTREAPBaseLeft->baseOrRoute.base;
      leftBCont->root = leftTREAPBaseRoot;
       
      TREAPBaseOrTREAPRouteNode * newRoute = treapAllocateRouteNode(splitKeyValueSize,
                                                                    splitKey,
                                                                    newTREAPBaseLeft,
                                                                    newTREAPBaseRight);
      //Link in new route
      *parent = newRoute;
      //Unlock threads waiting for old node
      node->valid = false;
      treapatree_unlock(&node->lock);
      //Done!
      enqueueInFreeList(set, baseNodeContainer);
#ifdef TREAPATREE_COUNT_ROUTE_NODES
      int count = atomic_fetch_add(&set->nrOfRouteNodes, 1);
      printf("%d\n", count);
#endif
      return true;
}

static inline TREAPBaseOrTREAPRouteNode * findLeftmostBaseNode(TREAPBaseOrTREAPRouteNode *  baseOrRouteParam,
                                                               TREAPBaseOrTREAPRouteNode ** parentRouteWriteBack){
    TREAPBaseOrTREAPRouteNode *  baseOrRoute = baseOrRouteParam; 
    TREAPBaseOrTREAPRouteNode *  prevNode = NULL;
    while(!baseOrRoute->isTREAPBaseNode){
        prevNode = baseOrRoute;
        baseOrRoute = baseOrRoute->baseOrRoute.route.left;
    }
    *parentRouteWriteBack = prevNode;
    return baseOrRoute;
}

static inline TREAPBaseOrTREAPRouteNode * findRightmostBaseNode(TREAPBaseOrTREAPRouteNode *  baseOrRouteParam,
                                                               TREAPBaseOrTREAPRouteNode ** parentRouteWriteBack){
    TREAPBaseOrTREAPRouteNode *  baseOrRoute = baseOrRouteParam; 
    TREAPBaseOrTREAPRouteNode *  prevNode = NULL;
    while(!baseOrRoute->isTREAPBaseNode){
        prevNode = baseOrRoute;
        baseOrRoute = baseOrRoute->baseOrRoute.route.right;
    }
    *parentRouteWriteBack = prevNode;
    return baseOrRoute;
}

static inline TREAPValueNode * takeOutRightmostValueNode(TREAPValueNode ** baseValueNodeParam){
    TREAPValueNode ** baseValueNode = baseValueNodeParam; 
    while((*baseValueNode)->right != NULL){
        baseValueNode = &(*baseValueNode)->right;
    }
    TREAPValueNode * toReturn = (*baseValueNode);
    (*baseValueNode) = (*baseValueNode)->left;
    toReturn->left = NULL;
    return toReturn;
}

static inline TREAPValueNode * takeOutLeftmostValueNode(TREAPValueNode ** baseValueNodeParam){
    TREAPValueNode ** baseValueNode = baseValueNodeParam; 
    while((*baseValueNode)->left != NULL){
        baseValueNode = &(*baseValueNode)->left;
    }
    TREAPValueNode * toReturn = (*baseValueNode);
    (*baseValueNode) = (*baseValueNode)->right;
    toReturn->right = NULL;
    return toReturn;
}

static inline
void takeOutRoot(TREAPBaseNode * base,
                 void (*custom_free)(void * v, void * context),
                 void * freeContext){
    TREAPValueNode ** currentNodePointer = &base->root;
    TREAPValueNode * node = *currentNodePointer;
    while (node->left != NULL || node->right != NULL) {
        //Go down
        if(node->left == NULL){
            //Rotate left
            TREAPValueNode * moveDown = node->right;
            *currentNodePointer = moveDown;
            node->right = moveDown->left;
            moveDown->left = node;
            currentNodePointer = &(moveDown->left);
        } else if(node->right == NULL){
            //Rotate right
            TREAPValueNode * moveDown = node->left;
            *currentNodePointer = moveDown;
            node->left = moveDown->right;
            moveDown->right = node;
            currentNodePointer = &(moveDown->right);
        } else if(node->left->priority > node->right->priority){
            //Rotate right
            TREAPValueNode * moveDown = node->left;
            *currentNodePointer = moveDown;
            node->left = moveDown->right;
            moveDown->right = node;
            currentNodePointer = &(moveDown->right);
        } else {
            //Rotate left
            TREAPValueNode * moveDown = node->right;
            *currentNodePointer = moveDown;
            node->right = moveDown->left;
            moveDown->left = node;
            currentNodePointer = &(moveDown->left);
        }
        node = *currentNodePointer;
    }
    *currentNodePointer = NULL;
    custom_free(node, freeContext);
}

static inline
bool treapNoContentionMerge(TREAPBaseOrTREAPRouteNode *  baseNodeContainer,
                            TREAPBaseOrTREAPRouteNode * parentRoute,
                            TREAPATreeSet * set,
                            int keyPosition,
                            void * (*extract_key)(void * v, int keyPos),
                            bool (*less_than)(void * v1, void * v2),
                            void (*custom_free)(void * v, void * context),
                            void * freeContext) {
    TREAPBaseNode * node = &baseNodeContainer->baseOrRoute.base;
    if(parentRoute == set->root){
        node->lock.statistics = 0;
        return false;
    }
    TREAPRouteNode * parent = &parentRoute->baseOrRoute.route;
    TREAPBaseNode * neighbourBase;
    bool neighbourLeft = true;
    TREAPBaseOrTREAPRouteNode * neighbourBaseContainer;
    TREAPBaseOrTREAPRouteNode * neighbourBaseParentRoute = NULL;
    if(baseNodeContainer == parent->left){
        neighbourLeft = false;
        neighbourBaseContainer = findLeftmostBaseNode(parent->right, &neighbourBaseParentRoute);
    } else {
        neighbourBaseContainer = findRightmostBaseNode(parent->left, &neighbourBaseParentRoute);
    }
    if(neighbourBaseParentRoute == NULL){
        neighbourBaseParentRoute = parentRoute;
    }
    neighbourBase = &neighbourBaseContainer->baseOrRoute.base;
    if(!treapatree_trylock(&neighbourBase->lock)){
        node->lock.statistics = 0;
        return false; // Lets retry later
    }
    
    if(!neighbourBase->valid){
        treapatree_unlock(&neighbourBase->lock);
        node->lock.statistics = 0;
        return false; // Lets retry later
    }
    // Ready to do the merge
    TREAPBaseOrTREAPRouteNode * newNeighbourBaseContainer = allocateTREAPBaseNode();
    TREAPBaseNode * newNeighbourBase = &newNeighbourBaseContainer->baseOrRoute.base;
    if(neighbourBase->root == NULL){
        newNeighbourBase->root = node->root;
    }else if(node->root != NULL){
        char fakeValue[1];
        TREAPValueNode * newFakeRoot = TREAPATREE_NEW_VALUE_NODE(1, fakeValue, freeContext);
        if(neighbourLeft){
            newFakeRoot->left = neighbourBase->root;
            newFakeRoot->right = node->root;
        }else{
            newFakeRoot->right = neighbourBase->root;
            newFakeRoot->left = node->root;
        }
        newNeighbourBase->root = newFakeRoot;
        takeOutRoot(newNeighbourBase, custom_free, freeContext);
    }
    //Take out the node
    //Lock parent
    tatas_lock(parent->lock);
    //Find parent of parent and lock if posible
    TREAPBaseOrTREAPRouteNode * parentOfParentRoute = NULL;
    do{
        if(parentOfParentRoute != NULL){
            tatas_unlock(parentOfParentRoute->baseOrRoute.route.lock);
        }
        parentOfParentRoute = 
            findParentOf(parentRoute,
                         parent->value,
                         set,
                         keyPosition,
                         extract_key,
                         less_than);
        if(parentOfParentRoute == NULL){
            //Parent of parent is root, we are safe becaue it can't be deleted
            break;
        }
        tatas_lock(parentOfParentRoute->baseOrRoute.route.lock);
    }while(!parentOfParentRoute->baseOrRoute.route.valid);
    //Parent and parent of parent is safe
    //We can now unlink the parent route 
    TREAPBaseOrTREAPRouteNode ** parentInLinkPtrPtr = NULL;
    if(parentOfParentRoute == NULL){
        parentInLinkPtrPtr = &set->root;
    }else if(parentOfParentRoute->baseOrRoute.route.left == parentRoute){
        parentInLinkPtrPtr = (TREAPBaseOrTREAPRouteNode **)&parentOfParentRoute->baseOrRoute.route.left;
    }else{
        parentInLinkPtrPtr = (TREAPBaseOrTREAPRouteNode **)&parentOfParentRoute->baseOrRoute.route.right;
    }
    if(neighbourLeft){
        *parentInLinkPtrPtr = parent->left;
    }else{
        *parentInLinkPtrPtr = parent->right;
    }
    //Unlink should have happened YEY
    //Unlock the locks
    parent->valid = false;
    tatas_unlock(parent->lock);
    if(parentOfParentRoute != NULL){
        tatas_unlock(parentOfParentRoute->baseOrRoute.route.lock);
    }
    //Link in new neighbour base
    if(*parentInLinkPtrPtr == neighbourBaseContainer){
        *parentInLinkPtrPtr = newNeighbourBaseContainer;
    }if(neighbourBaseParentRoute->baseOrRoute.route.left == neighbourBaseContainer){
        neighbourBaseParentRoute->baseOrRoute.route.left = newNeighbourBaseContainer;
    }else{
        neighbourBaseParentRoute->baseOrRoute.route.right = newNeighbourBaseContainer;
    }
    neighbourBase->valid = false;
    treapatree_unlock(&neighbourBase->lock);
    node->valid = false;
    treapatree_unlock(&node->lock);
    enqueueInFreeList(set, baseNodeContainer);
    enqueueInFreeList(set, neighbourBaseContainer);
    enqueueInFreeList(set, parentRoute);
#ifdef TREAPATREE_COUNT_ROUTE_NODES
    int count = atomic_fetch_sub(&set->nrOfRouteNodes, 1);
    printf("%d\n", count);
#endif
   return true;
}

static inline char * my_to_string3(void * value){
    int  valuei = *((int *)value);
    char * str = malloc(100);
    sprintf(str, "\"P%d V%d\"", (int)*((int*)&((char*)value)[offsetof(TREAPValueNode, priority) - offsetof(TREAPValueNode, value)]), valuei);
    return str;
}

static inline
void tat_set_print_helper_helper_dot2(TREAPValueNode * node, char * (*to_string)(void * v1)){
    if(node!=NULL){
        if(node->left !=NULL){
            char * valueString = to_string(node->value);
            printf("%s", valueString);
            TREAPATREE_FREE(valueString);
            printf(" -> ");
            valueString = to_string(node->left->value);
            printf("%s", valueString);
            TREAPATREE_FREE(valueString);
            printf(";\n");
        }
        if(node->right !=NULL){
            char * valueString = to_string(node->value);
            printf("%s", valueString);
            TREAPATREE_FREE(valueString);
            printf(" -> ");
            valueString = to_string(node->right->value);
            printf("%s", valueString);
            TREAPATREE_FREE(valueString);
            printf(";\n");
        }
        if(node->left == node){
            printf("LEFT EXIT\n");
            exit(0);
        }

        if(node->right == node){
            printf("RIGHT EXIT\n");
            exit(0);
        }
        tat_set_print_helper_helper_dot2(node->left, to_string);
        tat_set_print_helper_helper_dot2(node->right, to_string);
    }
}


static inline
bool tat_update_helper_TREAPValueNode(TREAPBaseNode * baseNode,
                                      void * valuePtr,
                                      unsigned int valueSize,
                                      int keyPosition,
                                      bool overwrite,
                                      void * (*extract_key)(void * v, int keyPos),
                                      bool (*are_equal)(void * v1, void * v2),
                                      bool (*less_than)(void * v1, void * v2),
                                      void ** replacedWriteBack,
                                      void (*custom_free)(void * v, void * context),
                                      void * freeContext,
                                      TREAPValueNode *** pointersStackParam,
                                      int * dirStackParam,
                                      int stackSizeParam
                                      ){
    int stackPos = 0;
    TREAPValueNode *** pointersStack;
    int * dirStack;
    int stackSize;
    TREAPValueNode **currentNodePointer = &baseNode->root;
    void * key;
    TREAPValueNode ** pointersStackS[TREAPATREE_DEFAULT_STACK_SIZE];
    int dirStackS[TREAPATREE_DEFAULT_STACK_SIZE];
    if(stackSizeParam == 0){
        stackSize = TREAPATREE_DEFAULT_STACK_SIZE;
        pointersStack = pointersStackS;
        dirStack = dirStackS;
    }else{
        stackSize = stackSizeParam;
        pointersStack = pointersStackParam;
        dirStack = dirStackParam;
    }

    key = TREAPATREE_EXTRACT_KEY_UNTRANSFORMED(valuePtr);

    dirStack[0] = TREAPATREE_ROOT_DIR;
    pointersStack[0] = currentNodePointer;
    while (true) {
	if(stackPos == (stackSize-1)){
            if(stackSizeParam != 0){
                free(dirStack);
                free(pointersStack);
            }
            stackSize = 2 * stackSize;
            pointersStack = malloc(stackSize * sizeof(TREAPValueNode **));
            dirStack = malloc(stackSize * sizeof(int));
            return tat_update_helper_TREAPValueNode(baseNode,
                                                    valuePtr,
                                                    valueSize,
                                                    keyPosition,
                                                    overwrite,
                                                    extract_key,
                                                    are_equal,
                                                    less_than,
                                                    replacedWriteBack,
                                                    custom_free,
                                                    freeContext,
                                                    pointersStack,
                                                    dirStack,
                                                    stackSize);
        }else if (*currentNodePointer == NULL) {
            *currentNodePointer = TREAPATREE_NEW_VALUE_NODE(valueSize, valuePtr, freeContext);
	    (*currentNodePointer)->priority = treap_random_number();
	    (*currentNodePointer)->left = NULL;
	    (*currentNodePointer)->right = NULL;
	    break;
	} else if (less_than(key, extract_key((*currentNodePointer)->value, keyPosition))) {
	    /* left */
            stackPos = stackPos + 1;
	    dirStack[stackPos] = TREAPATREE_LEFT_DIR;
	    currentNodePointer = &((*currentNodePointer)->left);
	    pointersStack[stackPos] = currentNodePointer;
	} else if (are_equal(key, extract_key((*currentNodePointer)->value, keyPosition))) { 
            if(overwrite){
                TREAPValueNode * old = *currentNodePointer;
                TREAPValueNode * oldLeft = old->left;
                TREAPValueNode * oldRight = old->right;
                int oldPriority = old->priority;
                *currentNodePointer = TREAPATREE_REPLACE_VALUE_NODE(valueSize, old, valuePtr, custom_free, freeContext);
                (*currentNodePointer)->left = oldLeft;
                (*currentNodePointer)->right = oldRight;
                (*currentNodePointer)->priority = oldPriority;
                if(stackSizeParam != 0){
                    free(dirStack);
                    free(pointersStack);
                }
                return true;
            }else{
                if(stackSizeParam != 0){
                    free(dirStack);
                    free(pointersStack);
                }
                return false;
            }
	    break;
	} else { 
            /* right */
            stackPos = stackPos + 1;
	    dirStack[stackPos] = TREAPATREE_RIGHT_DIR;
	    currentNodePointer = &((*currentNodePointer)->right);
	    pointersStack[stackPos] = currentNodePointer;
	}
    }
    stackPos = stackPos -1;//Now points to the pointer to the parent of the inserted node
    TREAPValueNode * node = (*currentNodePointer);
    while (stackPos >= 0 && (node->priority) > (*pointersStack[stackPos])->priority) {
        if(dirStack[stackPos + 1] == TREAPATREE_LEFT_DIR){
            TREAPValueNode * rotateUp = *pointersStack[stackPos];
            *pointersStack[stackPos] = node;
            rotateUp->left = node->right;
            node->right = rotateUp;
        } else {
            TREAPValueNode * rotateUp = *pointersStack[stackPos];
            *pointersStack[stackPos] = node;
            rotateUp->right = node->left;
            node->left = rotateUp;
        } 
        stackPos = stackPos - 1;
    }
    if(stackSizeParam != 0){
        free(dirStack);
        free(pointersStack);
    }
    return true;
}

static inline
bool tat_set_insert_opt(TREAPATreeSet * set,
                       void * valuePtr,
                       unsigned int valueSize,
                       int keyPosition,
                       bool overwrite,
                       void * (*extract_key)(void * v, int keyPos),
                       bool (*are_equal)(void * v1, void * v2),
                       bool (*less_than)(void * v1, void * v2),
                       bool copyIn,
                       void ** replacedWriteBack,
                       void (*custom_free)(void * v, void * context),
                       void * freeContext){
    void * key = TREAPATREE_EXTRACT_KEY_UNTRANSFORMED(valuePtr);
    //Find base node
    TREAPBaseOrTREAPRouteNode * currentNode = set->root;
    TREAPBaseOrTREAPRouteNode * prevNode = set->root;
    TREAPRouteNode * currentTREAPRouteNode = NULL;
    while(!currentNode->isTREAPBaseNode){
        currentTREAPRouteNode = &currentNode->baseOrRoute.route;
        void * nodeKey = extract_key(currentTREAPRouteNode->value, keyPosition);
        prevNode = currentNode;
        if(less_than(key, nodeKey)) {
            currentNode = currentTREAPRouteNode->left;
        } else {
            currentNode = currentTREAPRouteNode->right;
        }
    }
    //Lock and handle contention
    TREAPBaseNode * node = &currentNode->baseOrRoute.base;
    TreapatreeLock * lock = &node->lock;
    void ** parent = NULL;
    bool contention = true;
    treapatree_lock(lock);
    if(!node->valid){
        //Retry
        treapatree_unlock(lock);
        drmcs_runlock(&set->globalLock);
        drmcs_rlock(&set->globalLock);
        return tat_set_insert_opt(set,
                                 valuePtr,
                                 valueSize,
                                 keyPosition,
                                 overwrite,
                                 extract_key,
                                 are_equal,
                                 less_than,
                                 copyIn,
                                 replacedWriteBack,
                                 custom_free,
                                 freeContext);
    } else if(lock->statistics > TREAPATREE_MAX_CONTENTION_STATISTICS){
        if(currentTREAPRouteNode == NULL){
            parent = (void**)&set->root;
        }else if(currentTREAPRouteNode->left == currentNode){
            parent = (void**)&currentTREAPRouteNode->left;
        }else{
            parent = (void**)&currentTREAPRouteNode->right;
        }
    } else if(lock->statistics < TREAPATREE_MIN_CONTENTION_STATISTICS){
        if(currentTREAPRouteNode == NULL){
            parent = (void**)set->root;
        }else if(currentTREAPRouteNode->left == currentNode){
            parent = (void**)currentTREAPRouteNode->left;
        }else{
            parent = (void**)currentTREAPRouteNode->right;
        }
        contention = false;
    }
    //Handle empty before and then normal case
    bool returnValue = false;
    if(node->root == NULL) {
        TREAPValueNode * newNode = TREAPATREE_NEW_VALUE_NODE(valueSize, valuePtr, freeContext);
        newNode->priority = treap_random_number();
        node->root = newNode;
        returnValue = true;
    } else {
        returnValue = tat_update_helper_TREAPValueNode(node,
                                                       valuePtr,
                                                       valueSize,
                                                       keyPosition,
                                                       overwrite,
                                                       extract_key,
                                                       are_equal,
                                                       less_than,
                                                       replacedWriteBack,
                                                       custom_free,
                                                       freeContext,
                                                       NULL,
                                                       NULL,
                                                       0);
    }
    if(parent != NULL){
        if(contention){
            if(!treapContentionReduceSplit(currentNode,
                                           parent,
                                           set,
                                           keyPosition,
                                           extract_key,
                                           less_than)){
                treapatree_unlock(lock);
            }
        } else {
            if(!treapNoContentionMerge(currentNode,
                                       prevNode,
                                       set,
                                       keyPosition,
                                       extract_key,
                                       less_than,
                                       custom_free,
                                       freeContext)){
                treapatree_unlock(lock);
            }
        }
        return returnValue;
    }
    treapatree_unlock(lock);
    return returnValue;
}

static inline
void tat_set_insert(void * setParam, void * valuePtr, unsigned int valueSize){
    TREAPATreeSet * set = (TREAPATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    tat_set_insert_opt(set,
                      valuePtr,
                      valueSize,
                      set->keyPosition,
                      true,
                      set->extract_key,
                      set->are_equal,
                      set->less_than,
                      true,
                      NULL,
                      set->free,
                      set->freeContext);
    drmcs_runlock(&set->globalLock);
}

static inline
void tat_set_insert_ops(void * setParam, void * valuePtr, unsigned int valueSize, bool copyIn, void ** replacedWriteBack){
    TREAPATreeSet * set = (TREAPATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    tat_set_insert_opt(set,
                      valuePtr,
                      valueSize,
                      set->keyPosition,
                      true,
                      set->extract_key,
                      set->are_equal,
                      set->less_than,
                      copyIn,
                      replacedWriteBack,
                      set->free,
                      set->freeContext);
    drmcs_runlock(&set->globalLock);
}



static inline
bool tat_set_insert_new(void * setParam, void * valuePtr, unsigned int valueSize){
    TREAPATreeSet * set = (TREAPATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    bool returnValue = tat_set_insert_opt(set,
                                         valuePtr, valueSize,
                                         set->keyPosition,
                                         false,
                                         set->extract_key,
                                         set->are_equal,
                                         set->less_than,
                                         true,
                                         NULL,
                                         set->free,
                                         set->freeContext);
    drmcs_runlock(&set->globalLock);
    return returnValue;
}


#ifndef TREAPATREE_SPECIAL_COPY_OUT
#    define TREAPATREE_SPECIAL_COPY_OUT_PARAMS
#    define TREAPATREE_SPECIAL_COPY_OUT_PARAMS_PASS
#    define TREAPATREE_SPECIAL_COPY_OUT_PARAMS_PASS_NULL
#endif

static inline
void * tat_lookup_TREAPValueNode(TREAPValueNode * searchNode,
                           void * key,
                           int keyPosition,
                           void * (*extract_key)(void * v, int keyPos),
                           bool (*are_equal)(void * v1, void * v2),
                           bool (*less_than)(void * v1, void * v2) TREAPATREE_SPECIAL_COPY_OUT_PARAMS){
    TREAPValueNode * currentNode = searchNode;
    while(currentNode != NULL){
        void * nodeKey = extract_key(currentNode->value, keyPosition);
        if(less_than(key, nodeKey)) {
            currentNode = currentNode->left;
        } else if (are_equal(key, nodeKey)) {
#ifdef TREAPATREE_SPECIAL_COPY_OUT
                TREAPATREE_SPECIAL_COPY_OUT
#else
                void * toReturn = TREAPATREE_MALLOC(currentNode->valueSize);
                memcpy(toReturn, currentNode->value, currentNode->valueSize);
#endif
                return toReturn;
        } else {
            currentNode = currentNode->right;
        }
    }
    return NULL;
}



static inline
void * tat_set_lookup_opt(TREAPATreeSet * set, 
                          void * key,
                          int keyPosition,
                          void * (*extract_key)(void * v, int keyPos),
                          bool (*are_equal)(void * v1, void * v2),
                          bool (*less_than)(void * v1, void * v2),
                          void (*custom_free)(void * v, void * context),
                          void * freeContext TREAPATREE_SPECIAL_COPY_OUT_PARAMS){
    //Find base node
    TREAPBaseOrTREAPRouteNode * currentNode = set->root;
    TREAPBaseOrTREAPRouteNode * prevNode = set->root;
    TREAPRouteNode * currentTREAPRouteNode = NULL;
    while(!currentNode->isTREAPBaseNode){
        currentTREAPRouteNode = &currentNode->baseOrRoute.route;
        void * nodeKey = extract_key(currentTREAPRouteNode->value, keyPosition);
        prevNode = currentNode;
        if(less_than(key, nodeKey)) {
            currentNode = currentTREAPRouteNode->left;
        } else {
            currentNode = currentTREAPRouteNode->right;
        }
    }
    //Lock and handle contention
    TREAPBaseNode * node = &currentNode->baseOrRoute.base;
    TreapatreeLock * lock = &node->lock;
    void ** parent = NULL;
    bool contention = true;
    treapatree_lock(lock);
    if(!node->valid){
        //Retry
        treapatree_unlock(lock);
        drmcs_runlock(&set->globalLock);
        drmcs_rlock(&set->globalLock);
        return tat_set_lookup_opt(set,
                                  key,
                                  keyPosition,
                                  extract_key,
                                  are_equal,
                                  less_than,
                                  custom_free,
                                  freeContext TREAPATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
    } else if(lock->statistics > TREAPATREE_MAX_CONTENTION_STATISTICS){
        if(currentTREAPRouteNode == NULL){
            parent = (void**)&set->root;
        }else if(currentTREAPRouteNode->left == currentNode){
            parent = (void**)&currentTREAPRouteNode->left;
        }else{
            parent = (void**)&currentTREAPRouteNode->right;
        }
    } else if(lock->statistics < TREAPATREE_MIN_CONTENTION_STATISTICS){
        if(currentTREAPRouteNode == NULL){
            parent = (void**)set->root;
        }else if(currentTREAPRouteNode->left == currentNode){
            parent = (void**)currentTREAPRouteNode->left;
        }else{
            parent = (void**)currentTREAPRouteNode->right;
        }
        contention = false;
    }
    //Handle empty before and then normal case
    void * returnValue = NULL;
    if(node->root == NULL) {
        returnValue = NULL;
    } else {
        returnValue = tat_lookup_TREAPValueNode(node->root,
                                          key,
                                          keyPosition,
                                          extract_key,
                                          are_equal,
                                          less_than TREAPATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
    }
    if(parent != NULL){
        if(contention){
            if(!treapContentionReduceSplit(currentNode,
                                           parent,
                                           set,
                                           keyPosition,
                                           extract_key,
                                           less_than)){
                treapatree_unlock(lock);
            }
        }else{
            if(!treapNoContentionMerge(currentNode,
                                       prevNode,
                                       set,
                                       keyPosition,
                                       extract_key,
                                       less_than,
                                       custom_free,
                                       freeContext)){
                treapatree_unlock(lock);
            }
        }
        return returnValue;
    }
    treapatree_unlock(lock);
    return returnValue;
}

static inline
void * tat_set_lookup(void * setParam, 
                     void * key){
    TREAPATreeSet * set = (TREAPATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    void * returnValue = tat_set_lookup_opt(set,
                                            key,
                                            set->keyPosition,
                                            set->extract_key,
                                            set->are_equal,
                                            set->less_than,
                                            set->free,
                                            set->freeContext TREAPATREE_SPECIAL_COPY_OUT_PARAMS_PASS_NULL);
    drmcs_runlock(&set->globalLock);
    return returnValue;
}

static inline
void * tat_set_lookup_ops(void * setParam, 
                         void * key TREAPATREE_SPECIAL_COPY_OUT_PARAMS){
    TREAPATreeSet * set = (TREAPATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    void * returnValue = tat_set_lookup_opt(set,
                                            key,
                                            set->keyPosition,
                                            set->extract_key,
                                            set->are_equal,
                                            set->less_than,
                                            set->free,
                                            set->freeContext TREAPATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
    drmcs_runlock(&set->globalLock);
    return returnValue;
}



static inline
void tat_set_delete_TREAPValueNode(TREAPBaseNode * baseNode,
                                 void * key,
                                 int keyPosition,
                                 void * (*extract_key)(void * v, int keyPos),
                                 bool (*are_equal)(void * v1, void * v2),
                                 bool (*less_than)(void * v1, void * v2),
                                 void (*custom_free)(void * v, void * context),
                                 void * freeContext){
    TREAPValueNode ** currentNodePointer = &baseNode->root;
    while (true) {
        if (*currentNodePointer == NULL) {
            return; //Fail to find node
	} else if (less_than(key, extract_key((*currentNodePointer)->value, keyPosition))) {
	    /* left */
	    currentNodePointer = &((*currentNodePointer)->left);
	} else if (are_equal(key, extract_key((*currentNodePointer)->value, keyPosition))) { 
            //Success found node to delete
	    break;
	} else { 
            /* right */
	    currentNodePointer = &((*currentNodePointer)->right);
	}
    }
    TREAPValueNode * node = *currentNodePointer;
    while (node->left != NULL || node->right != NULL) {
        //Go down
        if(node->left == NULL){
            //Rotate left
            TREAPValueNode * moveDown = node->right;
            *currentNodePointer = moveDown;
            node->right = moveDown->left;
            moveDown->left = node;
            currentNodePointer = &(moveDown->left);
        } else if(node->right == NULL){
            //Rotate right
            TREAPValueNode * moveDown = node->left;
            *currentNodePointer = moveDown;
            node->left = moveDown->right;
            moveDown->right = node;
            currentNodePointer = &(moveDown->right);
        } else if(node->left->priority > node->right->priority){
            //Rotate right
            TREAPValueNode * moveDown = node->left;
            *currentNodePointer = moveDown;
            node->left = moveDown->right;
            moveDown->right = node;
            currentNodePointer = &(moveDown->right);
        } else {
            //Rotate left
            TREAPValueNode * moveDown = node->right;
            *currentNodePointer = moveDown;
            node->right = moveDown->left;
            moveDown->left = node;
            currentNodePointer = &(moveDown->left);
        }
        node = *currentNodePointer;
    }
    *currentNodePointer = NULL;
    custom_free(node, freeContext);
    return;
}

static inline
void tat_set_delete_opt(TREAPATreeSet * set, 
                       void * key,
                       int keyPosition,
                       void * (*extract_key)(void * v, int keyPos),
                       bool (*are_equal)(void * v1, void * v2),
                       bool (*less_than)(void * v1, void * v2),
                       void (*custom_free)(void * v, void * context),
                       void * freeContext){
    //Find base node
    TREAPBaseOrTREAPRouteNode * currentNode = set->root;
    TREAPBaseOrTREAPRouteNode * prevNode = set->root;
    TREAPRouteNode * currentTREAPRouteNode = NULL;
    while(!currentNode->isTREAPBaseNode){
        currentTREAPRouteNode = &currentNode->baseOrRoute.route;
        void * nodeKey = extract_key(currentTREAPRouteNode->value, keyPosition);
        prevNode = currentNode;
        if(less_than(key, nodeKey)) {
            currentNode = currentTREAPRouteNode->left;
        } else {
            currentNode = currentTREAPRouteNode->right;
        }
    }
    //Lock and handle contention
    TREAPBaseNode * node = &currentNode->baseOrRoute.base;
    TreapatreeLock * lock = &node->lock;
    void ** parent = NULL;
    bool contention = true;
    treapatree_lock(lock);
    if(!node->valid){
        //Retry
        treapatree_unlock(lock);
        drmcs_runlock(&set->globalLock);
        drmcs_rlock(&set->globalLock);
        tat_set_delete_opt(set, 
                          key,
                          keyPosition,
                          extract_key,
                          are_equal,
                          less_than,
                          custom_free,
                          freeContext);
        return;
    } else if(lock->statistics > TREAPATREE_MAX_CONTENTION_STATISTICS){
        if(currentTREAPRouteNode == NULL){
            parent = (void**)&set->root;
        }else if(currentTREAPRouteNode->left == currentNode){
            parent = (void**)&currentTREAPRouteNode->left;
        }else{
            parent = (void**)&currentTREAPRouteNode->right;
        }
    } else if(lock->statistics < TREAPATREE_MIN_CONTENTION_STATISTICS){
        if(currentTREAPRouteNode == NULL){
            parent = (void**)set->root;
        }else if(currentTREAPRouteNode->left == currentNode){
            parent = (void**)currentTREAPRouteNode->left;
        }else{
            parent = (void**)currentTREAPRouteNode->right;
        }
        contention = false;
    }
    //Handle empty before and then normal case
    if(node->root != NULL) {
        tat_set_delete_TREAPValueNode(node,
                                    key,
                                    keyPosition,
                                    extract_key,
                                    are_equal,
                                    less_than,
                                    custom_free,
                                    freeContext);
    }
    if(parent != NULL){
        if(contention){
            if(!treapContentionReduceSplit(currentNode,
                                           parent,
                                           set,
                                           keyPosition,
                                           extract_key,
                                           less_than)){
                treapatree_unlock(lock);
            }
        }else{
            if(!treapNoContentionMerge(currentNode,
                                       prevNode,
                                       set,
                                       keyPosition,
                                       extract_key,
                                       less_than,
                                       custom_free,
                                       freeContext)){
                treapatree_unlock(lock);
            }
        }
        return;
    }
    treapatree_unlock(lock);
}

static inline
void tat_set_delete(void * setParam, 
                   void * key,
                   unsigned int keySize){
    (void)keySize;
    TREAPATreeSet * set = (TREAPATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    tat_set_delete_opt(set,
                      key,
                      set->keyPosition,
                      set->extract_key,
                      set->are_equal,
                      set->less_than,
                      set->free,
                      set->freeContext);
    drmcs_runlock(&set->globalLock);

}


static inline
void tat_set_free_helper_helper(TREAPValueNode * node,
                               void (*custom_free)(void * v, void * context),
                               void * freeContext){
    if(node!=NULL){
        tat_set_free_helper_helper(node->left, custom_free, freeContext);
        tat_set_free_helper_helper(node->right, custom_free, freeContext);
        custom_free(node, freeContext);
    }
}

static inline
void tat_set_free_helper(TREAPBaseOrTREAPRouteNode * root,
                        void (*custom_free)(void * v, void * freeContext),
                        void * freeContext){
    if(root->isTREAPBaseNode){
        tat_set_free_helper_helper(root->baseOrRoute.base.root, custom_free, freeContext);
        TREAPATREE_FREE(root);
    }else{
        TREAPRouteNode * route = &root->baseOrRoute.route;
        tat_set_free_helper(route->left, custom_free, freeContext);
        tat_set_free_helper(route->right, custom_free, freeContext);
        free(route->lock);
        TREAPATREE_FREE(root);
    }
}

static inline
void tat_set_destroy(void * setParam){
    TREAPATreeSet * set = (TREAPATreeSet *)setParam;
    for(int i = 0; i < TREAPATREE_FREE_LIST_SIZE; i++){
        TREAPBaseOrTREAPRouteNode * elementToFree = (TREAPBaseOrTREAPRouteNode *)atomic_load(&set->freeList[i]);
        if(elementToFree != NULL){
            if(elementToFree->isTREAPBaseNode){
                TREAPATREE_LOCK_DESTROY_FUNCTION(&elementToFree->baseOrRoute.base.lock.lock); //Free deleted base
                TREAPATREE_FREE(elementToFree);
            }else{
                free(elementToFree->baseOrRoute.route.lock); //Free deleted route
                TREAPATREE_FREE(elementToFree);   
            }
        }
    }
    tat_set_free_helper(set->root, set->free, set->freeContext);
}

static inline
void tat_set_free(void * setParam){
    tat_set_destroy(setParam);
    TREAPATREE_FREE(setParam);
}

static inline
unsigned int tat_set_size(TREAPATreeSet * set){
    (void)set;
    //TODO Fix code for this!
    return 0;
}


static inline
void tat_set_print_helper_helper(TREAPValueNode * node, char * (*to_string)(void * v1)){
    if(node==NULL){
        printf("NULL");
    }else{
        printf("V(");
        char * valueString = to_string(node->value);
        printf("%s", valueString);
        TREAPATREE_FREE(valueString);
        printf(",");
        tat_set_print_helper_helper(node->left, to_string);
        printf(",");
        tat_set_print_helper_helper(node->right, to_string);
        printf(")");
    }
}

static inline
void tat_set_print_helper(TREAPBaseOrTREAPRouteNode * root, char * (*to_string)(void * v1)){
    if(root->isTREAPBaseNode){
        printf("B(");
        tat_set_print_helper_helper(root->baseOrRoute.base.root, to_string);
        printf(")");
    }else{
        TREAPRouteNode * route = &root->baseOrRoute.route;
        printf("R(");
        char * valueString = to_string(route->value);
        printf("%s", valueString);
        TREAPATREE_FREE(valueString);
        printf(",");
        tat_set_print_helper(route->left, to_string);
        printf(",");
        tat_set_print_helper(route->right, to_string);
        printf(")");
    }
}

static inline
void tat_set_print_helper_helper_dot(TREAPValueNode * node, char * (*to_string)(void * v1)){
    if(node!=NULL){
        if(node->left !=NULL){
            char * valueString = to_string(node->value);
            printf("%s", valueString);
            TREAPATREE_FREE(valueString);
            printf(" -> ");
            valueString = to_string(node->left->value);
            printf("%s", valueString);
            TREAPATREE_FREE(valueString);
            printf(";\n");
        }
        if(node->right !=NULL){
            char * valueString = to_string(node->value);
            printf("%s", valueString);
            TREAPATREE_FREE(valueString);
            printf(" -> ");
            valueString = to_string(node->right->value);
            printf("%s", valueString);
            TREAPATREE_FREE(valueString);
            printf(";\n");
        }
        tat_set_print_helper_helper_dot(node->left, to_string);
        tat_set_print_helper_helper_dot(node->right, to_string);
    }
}

static inline
void tat_set_print_helper_dot(TREAPBaseOrTREAPRouteNode * root, char * (*to_string)(void * v1)){
    if(root->isTREAPBaseNode){
        printf("B(\n");
        printf("digraph G{\n");
        printf("  graph [ordering=\"out\"];\n");
        tat_set_print_helper_helper_dot(root->baseOrRoute.base.root, to_string);
        printf("}\n");
        printf("\n)");
    }else{
        TREAPRouteNode * route = &root->baseOrRoute.route;
        printf("R(");
        char * valueString = to_string(route->value);
        printf("%s", valueString);
        TREAPATREE_FREE(valueString);
        printf(",");
        tat_set_print_helper_dot(route->left, to_string);
        printf(",");
        tat_set_print_helper_dot(route->right, to_string);
        printf(")");
    }
}


static inline
void tat_set_print(void * setParam){
    TREAPATreeSet * set = (TREAPATreeSet *)setParam;
    tat_set_print_helper(set->root, set->to_string);
    printf("\n\nDOT:\n\n");
    tat_set_print_helper_dot(set->root, set->to_string);
    printf("\n\n");
}

static inline
char * tat_set_to_string(void * setParam){
    (void)setParam;
    //DODO FIX CODE FOR THIS!
    return NULL;
}

static inline
bool tat_set_is_concurrent(){
    return true;
}

static inline
OOSet * oo_tat_set_create(unsigned int keyPosition,
                         void * (*extract_key)(void * v, int keyPos),
                         unsigned int (*hash_key)(void * k),
                         bool (*are_equal)(void * v1, void * v2),
                         bool (*less_than)(void * v1, void * v2),
                         char * (*to_string)(void * v1)){
    TREAPATreeSet * set = plain_tat_set_create(keyPosition,
                                         extract_key,
                                         hash_key,
                                         are_equal,
                                         less_than,
                                         to_string);
    OOSet * ooset = TREAPATREE_MALLOC(sizeof(OOSet));
    ooset->set = set;
    ooset->m = &TREAPATREE_SET_METHOD_TABLE;
    return ooset;
}


#endif
